from __future__ import print_function
from argtools import command, argument
import logging
import pysam
from operator import attrgetter
from ..samutil import ReadCountGenerator, sam_intervals
from ..interval import Interval, parse_bed


@command.add_sub
@argument('sam')
@argument.exclusive(
  argument('-r', '--regions', nargs='*', help='region e.g. one based indexes chr1, chr1:1001, or chr1:1001-2000'),
  argument('-b', '--bed', help='region defined bed file'))
@argument('-w', '--window', type=int, default=50, help='width of counter')
@argument('--offset', type=int, default=0, help='offset of regions (or bed file)')
@argument('-H', '--no-header', dest='header', action='store_false', default=True, help='add header')
@argument('--skip-flag', type=lambda x: int(x, 16), default=0x904)
@argument('-m', '--mass', choices=['middle', 'start', 'end', 'overlap'], default='middle', help='mass points of read used to determine the read is inside of the window')
@argument('--no-mate', dest='use_mate', action='store_false', default=True)
def bam_read_counter(args):
    sam = pysam.Samfile(args.sam)

    with sam:
        attrs = ['start', 'end', 'count', 'mapq1', 'clip', 'bclip']
        if args.use_mate:
            mate_info = ['unmapped', 'jumped', 'far', 'overlap', 'penetrate', 'rr', 'ff']
            attrs += ['mate.{0}'.format(m) for m in mate_info]
        attrs += ['covlen', 'covlen_mapq1']

        if args.header:
            print ('chrom', *attrs, sep='\t')

        if args.bed:
            it = parse_bed(args.bed, offset=args.offset)
        else:
            it = sam_intervals(sam, args.regions)
        for (chrom, start, end, _) in it:
            counters = ReadCountGenerator(sam, chrom, start=start, end=end, window=args.window, skip_flag=args.skip_flag, mass=args.mass)

            getter = attrgetter(*attrs)
            try:
                for counter in counters:
                    print (chrom, *getter(counter), sep='\t')
            except Exception as e:
                logging.error(e)

@command.add_sub
@argument('sam')
@argument.exclusive(
  argument('-r', '--regions', nargs='*', help='regions e.g. one based indexes chr1, chr1:1001, or chr1:1001-2000'),
  argument('-b', '--bed', help='region defined bed file'))
@argument('-w', '--window', type=int, default=50, help='width of counter')
@argument('--offset', type=int, default=0, help='offset of regions (or bed file)')
@argument('--skip-flag', type=lambda x: int(x, 16), default=0x904)
@argument('--row', type=int, default=3)
@argument('--col', type=int, default=2)
@argument('--use-name', action='store_true')
@argument('--name-col', type=int, default=4)
@argument('-o', '--output', default='output.pdf')
@argument('--font-size', type=float, default=20.)
@argument('--min-count', type=int, default=0)
def bam_read_count_viz(args):
    sam = pysam.Samfile(args.sam)
    window = args.window
    logging.info('Output : %s', args.output)
    logging.info('Window size : %s', window)

    import pandas as pd
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.cm as cm
    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_pdf import PdfPages
    import seaborn
    def vline(x, *args, **kwds):
        plt.gca().add_line(matplotlib.lines.Line2D([x, x], plt.gca().get_ylim(), *args, **kwds))

    def hline(y, *args, **kwds):
        plt.gca().add_line(matplotlib.lines.Line2D(plt.gca().get_xlim(),[y, y], *args, **kwds))

    def subplot_gen(row, col, on_page_change=None):
        page = 1
        while 1:
            if page > 1 and on_page_change:
                on_page_change()
            plt.figure()
            for i in xrange(row * col):
                yield page, plt.subplot(row, col, i+1)
            page += 1

    def on_page_change():
        pp.savefig(plt.gcf())

    with sam:
        attrs = ['start', 'end', 'count', 'mapq1', 'clip', 'bclip']
        mate_info = ['unmapped', 'jumped', 'far', 'overlap', 'penetrate', 'rr', 'ff']
        attrs += ['mate.{0}'.format(m) for m in mate_info]
        attrs += ['covlen', 'covlen_mapq1']
        getter = attrgetter(*attrs)

        with PdfPages(args.output) as pp:
            zoom = 1
            matplotlib.rcParams['figure.figsize'] = (11.7 * zoom, 8.3 * zoom)
            matplotlib.rcParams['font.size'] = args.font_size
            matplotlib.rcParams['axes.titlesize'] = 'medium'
            matplotlib.rcParams['axes.labelsize'] = 'medium'
            matplotlib.rcParams['axes.labelpad'] = 8.
            matplotlib.rcParams['xtick.labelsize'] = 'small'
            matplotlib.rcParams['ytick.labelsize'] = 'small'
            matplotlib.rcParams['legend.fontsize'] = 'small'

            if args.bed:
                it = parse_bed(args.bed, offset=args.offset)
            else:
                it = sam_intervals(sam, args.regions)
            subplot = subplot_gen(args.row, args.col, on_page_change=on_page_change)
            for (chrom, start, end, data) in it:
                logging.info('Plotting %s:%s-%s', chrom, start+1, end)
                counters = ReadCountGenerator(sam, chrom, start=start, end=end, window=args.window, skip_flag=args.skip_flag, mass='overlap')
                tab = pd.DataFrame([getter(counter) for counter in counters], columns=attrs)
                tab['mid'] = (tab['start'] + tab['end']) / 2.
                tab['density'] = 1. * tab['covlen'] / window   # like depth
                tab['density_mapq1'] = 1. * tab['covlen_mapq1'] / window   # like depth

                density_mean = tab['density'].mean()  # like depth
                logging.debug('\n%s', tab.head())
                logging.debug('\n%s', tab.tail())

                if len(tab) < args.min_count:  # skip plotting
                    logging.info('skipped')
                    continue

                page, ax = next(subplot)
                plt.tight_layout()
                #plt.title('{0}:{1}-{2}', chrom, start, end)
                #ax = plt.gca()
                if len(tab):
                    tab.plot(kind='area', x='mid', y='density', ax=ax)
                    tab.plot(kind='area', x='mid', y='density_mapq1', ax=ax)
                if start is not None and end is not None:
                    ax.set_xlabel('{0}:{1}-{2} ({3})'.format(chrom, start+1, end, (end - start)))
                else:
                    ax.set_xlabel('{0}'.format(chrom))
                hline(density_mean, color='red')
                if args.use_name:
                    title = data[args.name_col - 1]
                    plt.title(title)
            on_page_change()



import subprocess
import logging
import sys
import shlex
from six import string_types
try:
    from shlex import quote
except ImportError:
    from pipes import quote

CalledProcessError = subprocess.CalledProcessError
dry_run = False
default_shell = '/bin/bash'
default_shell_mode = True

# sync api
def call(args, **kwds):
    stdout = kwds.pop('stdout', sys.stdout)
    _dry_run = kwds.get('dry_run', dry_run)
    if not _dry_run and isinstance(stdout, string_types):
        with open(stdout, 'w+') as stdout_fp:
            logging.debug('Opend for stdout: %s', stdout)
            ret = call(args, stdout=stdout_fp, **kwds)
        logging.debug('Closed for stdout: %s', stdout)
        return ret

    stderr = kwds.pop('stderr', sys.stderr)
    #stderr = kwds.pop('stderr', subprocess.PIPE)
    if not _dry_run and isinstance(stderr, string_types):
        with open(stderr, 'w+') as stderr_fp:
            logging.debug('Opened for stderr: %s', stderr)
            ret = call(args, stderr=stderr_fp, **kwds)
        logging.debug('Closed for stderr: %s', stderr)
        return ret

    kwds.setdefault('executable', default_shell)
    kwds.setdefault('shell', default_shell_mode)
    if isinstance(args, string_types):
        cmd = args
    else:
        args = map(str, args)
        cmd = ' '.join(map(quote, args))
        #args = shlex.split(args[0])
    logging.info('$ %s', cmd)
    #logging.debug('$ %s', ' '.join(map(quote, args)))
    if not _dry_run:
        kwds.pop('dry_run', None)

        # adhoc
        # stdout_is_buffer = not hasattr(stdout, 'fileno')
        # stderr_is_buffer = not hasattr(stderr, 'fileno')
        # if stdout_is_buffer:
        #     stdout_buffer = stdout
        #     stdout = subprocess.PIPE
        # if stderr_is_buffer:
        #     stderr_buffer = stderr
        #     stderr = subprocess.PIPE
        p = subprocess.Popen(cmd, stdout=stdout, stderr=stderr, **kwds)
        logging.info('pid: %s starts', p.pid)
        # if stdout_is_buffer:
        #     stdout_buffer.write(p.stdout.read())
        # if stderr_is_buffer:
        #     stderr_buffer.write(p.stderr.read())
        ret_code = p.wait()
        logging.info('pid: %s ends with returncode: %s', p.pid, ret_code)
        if ret_code != 0:
            raise CalledProcessError(ret_code, cmd)
    #return p  # TODO change to return code ?



#!/usr/bin/env python
from __future__ import print_function
from argtools import command, argument
import logging
import numpy as np
import itertools
from collections import namedtuple, defaultdict
from itertools import groupby
from builtins import zip
from multiprocess import Pool
import signal
from scipy.stats import norm
from ..covfit import load_gender_info, get_ploidies
#from ..covfit import gmm
from ..covfit import gmm_v2 as gmm
from ..utils import iter_tabs, Counter, with_header, attrdict


_InputRow = namedtuple('_InputRow', 'chrom start end region_id samples covs valid_ratio valid_len'.split(' '))

def gen_covfit_inputs(args):
    if args.gender_info:
        sample_genders = load_gender_info(args.gender_info)
    else:
        sample_genders = defaultdict(lambda : 1)
    zoom = 2. / args.diploid_value
    max_copy = getattr(args, 'max_copy', None)

    if args.format == 'vertical':
        def iter_inputs(it):
            for region_id, rows in groupby(it, lambda x: x[idx_region_id]):
                rows = list(rows)
                chrom = rows[0][0]
                start = int(rows[0][1]),
                end = int(rows[0][2]),
                samples = np.array([row[idx_sample_id] for row in rows])
                covs = np.array([row[idx_cov] for row in rows], dtype=float)
                valid_ratio = np.array([row[idx_valid_ratio] for row in rows], dtype=float)
                valid_len = np.array([row[idx_valid_len] for row in rows], dtype=int)
                yield _InputRow(chrom=chrom, start=start, end=end, region_id=region_id, samples=samples, covs=covs, valid_ratio=valid_ratio, valid_len=valid_len)

    elif args.format == 'horizontal':
        def iter_inputs(it):
            for row in it:
                chrom = row[0]
                start = int(row[1])
                end = int(row[2])
                region_id = row[idx_region_id]
                samples = np.array(row[idx_sample_id].split(','))
                covs = np.array(row[idx_cov].split(','), dtype=float)
                valid_ratio = np.array(row[idx_valid_ratio].split(','), dtype=float)
                valid_len = np.array(row[idx_valid_len].split(','), dtype=int)
                yield _InputRow(chrom=chrom, start=start, end=end, region_id=region_id, samples=samples, covs=covs, valid_ratio=valid_ratio, valid_len=valid_len)

    else:
        raise NotImplementedError(args.format)

    with open(args.coverage_bed) as fp:
        it = iter_tabs(fp)
        header = next(it)
        idx_region_id = header.index(args.col_region_id) if args.col_region_id else 3
        idx_sample_id = header.index(args.col_sample_id)
        idx_cov = header.index(args.col_cov)
        idx_valid_ratio = header.index('valid_ratio')
        idx_valid_len = header.index('valid_len')

        for row in iter_inputs(it):
            covs = row.covs * zoom # zoom

            # filter condition
            sel = (~np.isnan(covs) & (covs <= args.max_copy))
            if getattr(args, 'min_valid_ratio', None):
                sel &= row.valid_ratio >= args.min_valid_ratio
            if getattr(args, 'min_valid_len', None):
                sel &= row.valid_len >= args.min_valid_len

            mod_covs = covs[sel]
            mod_samples = row.samples[sel]

            genders = [sample_genders[s] for s in mod_samples]
            ploidies = get_ploidies(row.chrom, genders)
            ploidy_counts = Counter(ploidies)
            data = attrdict(
                    row = row,
                    mod_covs = mod_covs,
                    mod_samples = mod_samples,
                    args = args,
                    ploidies = ploidies,
                    nsamples = ','.join([str(ploidy_counts[cn]) for cn in (0, 1, 2)]),
            )
            yield data


@command.add_sub
@argument('coverage_bed')
@argument('--col-region-id')
@argument('--col-sample-id', default='sampleid')
@argument('--col-cov', default='mean_cov')
@argument('-F', '--format', default='vertical', choices=['vertical', 'horizontal'])
@argument('-d', '--diploid-value', type=float, default=1.0)
@argument('-m', '--max-copy', type=int, default=10)
@argument('-la', type=float, default=0.3, help='penalty of offset')
@argument('-lb', type=float, default=1.0, help='penalty of ratio')
@argument('-lc', type=float, default=0.3, help='penalty of precision')
@argument('-a', nargs='*', type=float, default=[0., 0.4, 0.8], help='penalty of offset')
@argument('-b', nargs='*', type=float, default=[1.0], help='penalty of ratio')
@argument('-c', type=float, default=1.0, help='penalty of precision')
@argument('-f', '--max-diff', type=float, default=1e-6, help='max diff of log likelihood')
@argument('-p', '--process', type=int, default=8)
@argument('--delta', type=float, default=0.3, help='variance for zero copy')
@argument('--step-limit', type=int, help='the number of steps until inference stopped')
@argument('--gender-info', help='col1: sample_id, col2: {1:male, 2:female}')
#TODO maybe min_valid_ratio and min_valid_len should be add to the options
def covfit_gmm(args):
    """

    1. chrom
    2. start
    3. end
    4. region id
    """
    # not required
    def imap_wrap(pool, task, task_inputs, chunksize=1):  # test
        result = pool.imap(task, task_inputs, chunksize=1)
        while 1:
            yield result.next(timeout=999999)


    def run():
        def init_pool():
            logging.info('Init pool')
            signal.signal(signal.SIGINT, signal.SIG_IGN)

        pool = Pool(args.process, init_pool)
        try:
            task_inputs = gen_covfit_inputs(args)

            if args.process == 1:
                results = itertools.imap(task, task_inputs)
            else:
                results = pool.imap(task, task_inputs, 1)
                #results = imap_wrap(pool, task, task_inputs, 100)

            row_header = ['chrom', 'start', 'end', 'region_id']
            data_header = ['nsamples']
            param_header = ['status', 'ia', 'ib', 'ic', 'a', 'b', 'c', 'top_cn', 'top_theta', 'theta', 'll', 'step']
            out_header = row_header + data_header + param_header
            print (*out_header, sep='\t')
            for data, params in results:
                row = [getattr(data.row, c) for c in row_header] \
                    + [getattr(data, c) for c in data_header] \
                    + [params[c] for c in param_header]
                print (*row, sep='\t')
            pool.close()
        except Exception as e:
            pool.terminate()
            raise e
        finally:
            pool.join()

    def task(data):
        chrom = data.row.chrom
        start = data.row.start
        end = data.row.end
        name = data.row.region_id
        covs = data.mod_covs
        la = data.args.la
        lb = data.args.lb
        lc = data.args.lc
        init_as = data.args.a
        init_bs = data.args.b
        init_abs = [(a, b) for a in init_as for b in init_bs]
        init_c = data.args.c
        delta = data.args.delta
        max_diff = data.args.max_diff
        step_limit = data.args.step_limit
        max_copy = data.args.max_copy
        ploidies = data.ploidies

        #ploidy_counts = Counter(ploidies)  # 0, 1, 2
        #nsamples_str =  ','.join(str(ploidy_counts[i]) for i in (0, 1, 2))

        n_callable = len(covs)
        logging.info('[%s] n_callable: %s', name, n_callable)
        logging.debug('%s', data)
        #init_as = (0.,)
        results = []
        for init_a, init_b in init_abs:
            init_params = {
                    'status': 0,
                    'ia': init_a,
                    'ib': init_b,
                    'ic': init_c,
                    'a': init_a,
                    'b': init_b,
                    'c': init_c,
                    'theta': '',
                    'step': 0,
                    'top_cn': 0,
                    'top_theta': 0,
                    'll': float('-inf'),
            }
            norm = None
            try:
                assert n_callable > 0, 'The number of callable samples ({0}) is less than the minimum value'.format(n_callable)
                norm = gmm.Normalizer(covs, ploidies=ploidies, max_copy=max_copy, a=init_a, b=init_b, c=init_c, la=la, lb=lb, lc=lc, max_diff=max_diff, step_limit=step_limit, name=name)
                #init_a, init_b, _, _ = get_best_fit(mod_depths)
                result = norm.iteration()
                top_cn = np.argmax(result.theta)
                params = dict(init_params,
                        status=1, a=result.a, b=result.b, c=result.c,
                        theta = ','.join(map(str, result.theta)),
                        step = result.step,
                        top_cn = top_cn,
                        top_theta = result.theta[top_cn],
                        ll = result.ll,
                )
            except AssertionError as e:
                logging.warning('Assert error: %s', e)
                logging.warning('Cannot optimize model for %s', data)
                step = 0 if norm is None else norm.step
                params = dict(init_params, step=step)
            results.append(params)
        best_params = max(results, key=lambda v: (v['status'], v['ll']))

        logging.info('Emitting %s:%s-%s, %s, %s', chrom, start, end, name, best_params)
        return (data, best_params)

    run()


@command.add_sub
@argument('coverage_bed')
@argument('-p', '--fit-params', required=True)
@argument('-F', '--format', default='vertical', choices=['vertical', 'horizontal'])
@argument('--min-valid-ratio', type=float, default=.5)
@argument('--min-valid-len', type=int, default=100)
@argument('-t', '--threshold', type=float, default=.8)
@argument('--col-sample-id', default='sampleid')
@argument('--col-region-id')
@argument('--col-cov', default='mean_cov', help='column name of normalized coverage')
@argument('--delta', type=float, default=0.3, help='variance for zero copy')    # TODO incorporate into the result
@argument('-d', '--diploid-value', type=float, default=1.0)
@argument('-m', '--max-copy', type=int, default=10)
@argument('--gender-info', help='col1: sample_id, col2: {1:male, 2:female}')
def covfit_gmm_genotype(args):
    """
    gt        # filtered by specified condition
    map_gt    #
    likelihoods
    """
    fit_it = with_header(iter_tabs(open(args.fit_params)), use_header=True)
    fit_header = next(fit_it)
    #col_region_id = args.col_region_id if args.col_region_id else 'name'
    cov_header = 'chrom start end region_id'.split(' ')
    header = cov_header + ['samples', 'copy_number', 'gts', 'gt_probs']
    print (*header, sep='\t')
    covfit_inputs = gen_covfit_inputs(args)

    NA = float('nan')
    for data, params in zip(covfit_inputs, fit_it):
        chrom = data.row.chrom
        start = data.row.start
        end = data.row.end
        name = data.row.region_id
        assert (chrom, start, end, name) == (params['chrom'], int(params['start']), int(params['end']), params['region_id']), 'chrom, start, and end should be aligned for the two input files'
        call_covs = data.mod_covs
        call_samples = data.mod_samples
        ploidies = data.ploidies
        chrom = params['chrom']
        start = int(params['start'])
        end = int(params['end'])

        try:
            a = float(params['a'])
            b = float(params['b'])
            c = float(params['c'])
            theta = np.array(map(float, params['theta'].split(',')))
        except ValueError:
            theta = NA

        sample_gt_keys = {}
        sample_gt_probs = {}
        sample_top_cns = {}
        if sum(ploidies) > 0 and not np.any(np.isnan(theta)):
            norm = gmm.Normalizer(call_covs, ploidies=ploidies, max_copy=args.max_copy, a=a, b=b, c=c, theta=theta, name=name, delta=args.delta)
            cn_probs = np.array(list(norm.iter_cns()))
            top_cns = np.argmax(cn_probs, axis=1)
            top_cn_probs = np.array([cn_probs[i, c] for i, c in enumerate(top_cns)])
            logging.info(cn_probs)
            logging.info(top_cns)
            logging.info(top_cn_probs)
            called_sel = top_cn_probs >= args.threshold
            sample_top_cns.update(zip(call_samples[called_sel], top_cns[called_sel]))
            for sample, gts in zip(call_samples, norm.iter_genotypes()):
                gt_keys, gt_vals = get_allelotype_records(gts)
                sample_gt_keys[sample] = gt_keys
                sample_gt_probs[sample] = gt_vals

        top_cns = [str(sample_top_cns.get(s, '.')) for s in data.row.samples]
        gt_keys = [';'.join(map(str, sample_gt_keys.get(s, ()))) or '.' for s in data.row.samples]
        gt_probs = [';'.join(map(str, sample_gt_probs.get(s, ()))) or '.' for s in data.row.samples]

        out_row = list(getattr(data.row, col) for col in cov_header)
        out_row.append(','.join(data.row.samples))
        out_row.append(','.join(top_cns))
        out_row.append(','.join(gt_keys))
        out_row.append(','.join(gt_probs))
        print (*out_row, sep='\t')


@command.add_sub
@argument('coverage')
@argument('-p', '--fit-params', required=True)
@argument('--min-valid-ratio', type=float, default=.5)
@argument('--min-valid-len', type=int, default=100)
#@argument('--col-region-id')
@argument('--col-cov', default='mean_cov', help='column name of normalized coverage')
@argument('--delta', type=float, default=0.3, help='variance for zero copy')    # TODO incorporate into the result
@argument('-d', '--diploid-value', type=float, default=1.0)
@argument('-t', '--threshold', type=float, default=.8)
@argument('--sex', type=int, choices=[1, 2], required=True)
def covfit_gmm_genotype1(args):
    """
    gt        # filtered by specified condition
    map_gt    # 
    likelihoods
    """
    zoom = 2. / args.diploid_value
    cov_it = with_header(iter_tabs(open(args.coverage)), use_header=True)
    fit_it = with_header(iter_tabs(open(args.fit_params)), use_header=True)
    cov_header = next(cov_it)
    fit_header = next(fit_it)
    #col_region_id = args.col_region_id if args.col_region_id else 'name'
    header = list(cov_header) + ['copy_number', 'top_cn', 'top_gt', 'top_gt_prob', 'samples', 'gts', 'cn_probs']
    print (*header, sep='\t')

    NA = float('nan')
    for cov_row, params in zip(cov_it, fit_it):
        assert all(cov_row[col] == params[col] for col in ('chrom', 'start', 'end')), 'chrom, start, and end should be aligned for the two input files'
        chrom = params['chrom']
        start = int(params['start'])
        end = int(params['end'])
        name = params['region_id']

        depth = float(cov_row[args.col_cov])
        mod_depth = depth * zoom
        ploidy = get_ploidies(chrom, [args.sex])[0]

        try:
            a = float(params['a'])
            b = float(params['b'])
            c = float(params['c'])
            theta = np.array(map(float, params['theta'].split(',')))
        except ValueError:
            theta = NA

        cn_probs = (NA,)
        gt_keys = ['.']
        gt_vals = [1.]
        if ploidy == 0:
            top_cn = 0
            cn = 0
            gt_keys = ['0']
            gt_vals = [1.]
        elif np.any(np.isnan(theta)):
            max_copy = len(theta) - 1
            norm = gmm.Normalizer(np.array([mod_depth]), ploidies=[ploidy], max_copy=max_copy, a=a, b=b, c=c, theta=theta, name=name, delta=args.delta)
            cn_probs = np.array(list(norm.iter_cns()))
            top_cns = np.argmax(cn_probs, axis=1)
            cn = top_cn if cn_probs[top_cn] >= args.threshold else -1

            valid_ratio = float(cov_row['valid_ratio'])
            valid_len = int(cov_row['valid_len'])
            if valid_ratio < args.min_valid_ratio \
               or valid_len < args.min_valid_len:
                cn = -1


            gts = list(norm.iter_genotypes())[0]
            gt_keys, gt_vals = get_allelotype_records(gts)
            # get at records
        else:
            top_cn = -1
            cn = -1

        out_row = list(cov_row[col] for col in cov_header)
        out_row.append(cn)
        out_row.append(top_cn)
        out_row.append(gt_keys[0])
        out_row.append('{0:.4f}'.format(gt_vals[0]))
        out_row.append(':'.join(gt_keys))
        out_row.append(':'.join(map('{0:.4f}'.format, gt_vals)))
        out_row.append(':'.join(map('{0:.4f}'.format, cn_probs)))
        print (*out_row, sep='\t')


def get_allelotype_records(gts, psum_max=.99):
    """
    gts: {'0': 1.0}
         {'1/0': 0.8, '2/0': 0.1}
    """
    gt_keys = []
    gt_vals = []
    psum = .0
    for key, p in sorted(gts.items(), key=lambda x: x[1], reverse=True):
        gt_keys.append(key)
        gt_vals.append(p)
        psum += p
        if psum > psum_max:
            break
    return gt_keys, gt_vals



# -*- coding: utf-8 -*-
import logging
from operator import attrgetter
from .interval import Interval, IntervalMixin
import pysam
from .utils import fill_text, cached_property, attrdict
from builtins import zip


def parse_region(region):
    """
    Args:
        region: 1-based region
    Returns:
        (contig, start, end) 0-based coordinate
    """
    sp = region.split(':')
    contig = sp[0]
    if len(sp) == 1:
        return (contig, None, None)
    sp = sp[1].split('-')
    start = int(sp[0].replace(',', ''))
    if len(sp) == 1:
        return (contig, start - 1, None)
    return (contig, start - 1, int(sp[1].replace(',', '')))


def sam_intervals(sam, regions=None):
    """
    sam: pysam.Samfile
    """
    ref_lens = dict(zip(sam.references, sam.lengths))
    if regions is None:
        # ':' can be accidentally contained in refenrece names (e.g. HLA:HLA00001)
        for contig in sam.references:
            start = 0
            end = ref_lens[contig]
            yield Interval(contig, start, end, None)
    else:
        for r in regions:
            contig, start, end = parse_region(r)
            if start is None:
                start = 0
            if end is None:
                end = ref_lens[contig]
            yield Interval(contig, start, end, None)


import re
_blank_pat1 = re.compile('^\s+$')
def _is_blank(st):
    return bool(_blank_pat1.match(st))


# This does not work for multimapped records
class LocalInfo:
    def __init__(self, sam, iv, fasta=None, skip_flags=0x004):
        self.contig = iv.contig
        self.start = iv.start
        self.end = iv.end
        if fasta:
            fasta = pysam.Fastafile(fasta)
            ref_bases = fasta.fetch(reference=iv.contig, start=iv.start, end=iv.end).upper()
        else:
            ref_bases = '?' * (iv.end - iv.start)
        self._ref_bases = ref_bases
        self._reads = [Read(rec) for rec in sam.fetch(reference=iv.contig, start=iv.start, end=iv.end) if not rec.flag & skip_flags]

        self._bases_liss = bases_liss = []   # e.g. [['A', 'GC', '-', ...]]
        for read in self._reads:
            bases_lis = read.get_bases_list(iv.start, iv.end)
            bases_liss.append(bases_lis)

        pos_infos = []
        for bases_lis in zip(*bases_liss):
            length = max(1, max(len(b) for b in bases_lis))
            pos_infos.append({'length': length, 'bases': bases_lis})
        self._max_lens = lens = [r['length'] for r in pos_infos]
        self.align_length = sum(lens)

    def get_ref_seq(self):
        return ''.join(self._ref_bases)

    def get_ref_align(self):
        filled = [fill_text(bs, l, char='-') for l, bs in zip(self._max_lens, self._ref_bases)]   # fill missing bases
        return filled

    def iter_read_aligns(self):  # TODO sorted by position bases
        for read, bases_lis in zip(self._reads, self._bases_liss):
            filled = [fill_text(bs, l, char=(' ' if _is_blank(bs) else '-')) for l, bs in zip(self._max_lens, bases_lis)]   # fill missing bases
            yield (read, filled)

    def iter_align_pairs(self):
        """
        Yields: (Read, (aln, ref_aln))
        """
        ref_bases = self._ref_bases
        for read, bases_lis in zip(self._reads, self._bases_liss):
            b1s = []
            b2s = []
            for b1, b2 in zip(bases_lis, ref_bases):
                fill_len = max(len(b1), len(b2))
                #b1 = fill_text(b1, fill_len, char=(' ' if _is_blank(b1) else '-'))
                #b2 = fill_text(b2, fill_len, char=(' ' if _is_blank(b2) else '-'))
                b1 = fill_text(b1, fill_len, char=(' ' if _is_blank(b1) else '-'))
                b2 = fill_text(b2, fill_len, char=(' ' if _is_blank(b2) else '-'))
                b1s.append(b1)
                b2s.append(b2)
            seq1 = ''.join(b1s)
            seq2 = ''.join(b2s)
            yield read, (seq1, seq2)


def clip_seq(seq, cigar_str):
    """
    >>> clip_seq('ATCGATCG', '2M4I2M')
    'ATCGATCG'
    >>> clip_seq('ATCGATCG', '2S4I2S')
    'ATCGATCG'
    >>> clip_seq('ATCGATCG', '2H4I2H')  # only clip when hard clip
    'CGAT'
    """
    parsed = cigar_parse(cigar_str)
    if not parsed:
        return seq
    l = parsed[0][1] if parsed[0][0] == CIGAR.H else 0
    r = parsed[-1][1] if parsed[-1][0] == CIGAR.H else 0
    return seq[l:len(seq) - r]


class CIGAR:
    M = 0
    I = 1
    D = 2
    N = 3
    S = 4
    H = 5
    P = 6
    X = 7
    EQ = 8  # '='

_cigar_op_codes = dict((_op, _code) for _code, _op in enumerate('MIDNSHPX='))
_re_cigar = re.compile('(\d+)([MIDNSHPX=])')

def cigar_parse(cigar_str):
    """
    >>> cigar_parse('2I3M3D3M') == [(CIGAR.I, 2), (CIGAR.M, 3), (CIGAR.D, 3), (CIGAR.M, 3)]
    True
    >>> cigar_parse('3M2I3M2S') == [(CIGAR.M, 3), (CIGAR.I, 2), (CIGAR.M, 3), (CIGAR.S, 2)]
    True
    """
    ret = []
    for m in _re_cigar.finditer(cigar_str):
        l, op = m.groups()
        ret.append((_cigar_op_codes[op], int(l)))
    return ret


def alignment_to_cigar(aln, ref_aln):
    """
    >>> aln1 = '--------ATATGGGCCATCT'
    >>> aln2 = 'ATATATATATACGGG--ATAT'
    >>> alignment_to_cigar(aln1, aln2)
    [(2, 8), (0, 7), (1, 2), (0, 4)]
    """
    cigars = []
    op = None
    l = 0
    for a1, a2 in zip(aln, ref_aln):
        if a1 == '-':
            if op == CIGAR.D:
                l += 1
                continue
            if op is not None:
                cigars.append((op, l))
            op = CIGAR.D
            l = 1
        elif a2 == '-':
            if op == CIGAR.I:
                l += 1
                continue
            if op is not None:
                cigars.append((op, l))
            op = CIGAR.I
            l = 1
        else:
            if op == CIGAR.M:
                l += 1
                continue
            if op is not None:
                cigars.append((op, l))
            op = CIGAR.M
            l = 1
    cigars.append((op, l))
    return cigars


class Read(IntervalMixin):
    def __init__(self, rec):
        self._rec = rec
        self.name = rec.qname
        self.start = rec.pos
        self.end = rec.pos if rec.aend is None else rec.aend
        self.rec = rec   # pysam.AlignedRead
        self._seq = rec.seq
        self._bases = _get_read_bases(rec)

        self.has_lclip = (rec.qstart > 0)
        self.has_rclip = (rec.qend < rec.rlen)
        self.mapq = rec.mapq
        self.alen = 0 if self.end is None else self.end - self.start # aligned length
        self.tags = dict(rec.get_tags())
        self.edit = self.tags.get('NM', None)
        self.nins = 0
        self.ndel = 0
        for op, length in rec.cigar:
            if op == CIGAR.D:
                self.ndel += length
            elif op == CIGAR.I:
                self.nins += length

        #self.mismatches = 0
        self.qlen = rec.qlen
        self.unmapped = int(rec.is_unmapped)
        self.lclip = rec.qstart
        self.rclip = rec.rlen - rec.qend
        self.reverse = int(rec.is_reverse)
        self.suppl = int(rec.is_supplementary)
        self.read1 = int(rec.is_read1)
        self.contig = self.rname = self._rec.reference_name

        # set mate pattern
        # ---------->  ....   # right unmapped
        # ...   <----------  # left unmapped
        self.is_left = int(not rec.is_reverse)  # mate is expected to be in opposite side
        self.mate_miss = int(rec.mate_is_unmapped)
        self.mate_tid = rec.pnext
        self.tlen = tlen = rec.tlen
        # whether mate is far is determined from tlen and deviation of tlen distribution
        #self.mate_end = rec.pos + tlen if tlen > 0 else (rec.pos is rec.end is None else rec.aend) - tlen # is ok?
        self.mate_invert = 0
        self.mate_back = 0
        self.mate_jump = 0

        if self.mate_miss:
            return
        self.mate_jump = int(rec.tid != self.mate_tid)
        self.mate_invert = int(rec.mate_is_reverse == rec.is_reverse)
        if not rec.is_reverse:
            self.mate_back = int(rec.pnext < rec.pos)
        else:
            self.mate_back = int(rec.pos < rec.pnext)

    @cached_property
    def fragmented(self):
        return self.suppl or 'SA' in self.tags

    @cached_property
    def num_frags(self):
        sa = self.tags.get('SA')
        if sa is None:
            return 1
        else:
            return 1 + sa.count(';')

    @cached_property
    def edit_ratio(self):
        return 1. * self.edit / self.alen

    def get_bases_list(self, start, end):
        EMPTY = ' '
        start_offset = start - self.start
        end_offset = end - self.end
        bases = self._bases
        if start_offset > 0:
            bases = self._bases[start_offset:]
        else:
            bases = [EMPTY] * (-start_offset) + bases
        if end_offset < 0:
            bases = bases[:end_offset]
        else:
            bases = bases + [EMPTY] * end_offset
        return bases


def _get_read_bases(rec):
    '''
    Inputs:
        rec: SAM record
    Returns:
        [(A|T|C|G|-|*)+] at aligned positions

    >>> _get_read_bases(attrdict(seq='AATCAGTA', cigar=cigar_parse('2I3M3D3M')))
    ['T', 'C', 'A', '-', '-', '-', 'G', 'T', 'A']
    >>> _get_read_bases(attrdict(seq='AATCAGTACC', cigar=cigar_parse('3M2I3M2S')))
    ['A', 'A', 'TCA', 'G', 'T', 'A']
    '''
    seq = rec.seq
    # mask low qualities
    #seq = ''.join(s if q >= 10 else '.' for (s, q) in zip(seq, rec.qual))
    offset = 0
    bases = []
    for (op, length) in rec.cigar:   # TODO (show clip?)
        if op in (0, 7, 8):  # M, X, =
            bases.extend(list(seq[offset:offset + length]))
            offset += length
        elif op == 1:   # I
            if bases:   # inserted seq. at start position will be ignored
                bases[-1] += seq[offset:offset + length]  # add to prev sequence
            offset += length
        elif op == 2:   # D
            bases.extend(['-'] * length)
        elif op == 3:   # N
            bases.extend(['*'] * length)
        elif op == 4:   # S
            offset += length
        elif op == 5:   # H
            pass
        elif op == 6:   # P
            raise NotImplementedError
    return bases


class ReadCounterBase(object):
    __slots__ = ('start', 'end', 'window',)

    def __init__(self, start, window=1):
        self.start = start
        self.end = start + window

    def add(self, rec):
        """
        Add SAM Record
        """
        raise NotImplementedError


class ReadCounter(ReadCounterBase):
    def __init__(self, start, window):
        super(ReadCounter, self).__init__(start, window=window)
        self.count = 0       # qstart is contained
        self.clip = 0
        self.lclip = 0
        self.rclip = 0
        self.bclip = 0
        self.reverse = 0
        self.mapq1 = 0
        self.mate = MateInfoCounter()
        self.covlen = 0   # covered length within the bin
        self.covlen_mapq1 = 0

    def add(self, rec):
        self.count += 1
        is_lclip = (rec.qstart > 0)
        is_rclip = (rec.qend < rec.rlen)
        is_mapq1 = (rec.mapq <= 1)
        covlen = min(self.end, rec.aend) - max(self.start, rec.pos)
        self.clip += int(is_lclip or is_rclip)
        self.lclip += int(is_lclip)
        self.rclip += int(is_rclip)
        self.bclip += int(is_lclip and is_rclip)
        self.reverse += rec.is_reverse
        self.mapq1 += int(is_mapq1)
        self.mate.add(rec)
        self.covlen += covlen
        if is_mapq1:
            self.covlen_mapq1 += covlen

    def __str__(self):
        return '\t'.join((str(self.start), str(self.end), str(self.__dict__)))


class MateInfoCounter(ReadCounterBase):
    """ Count mate info
    """
    __slots__ = ('_tlen_far', 'unmapped', 'jumped', 'far', 'overlap', 'penetrate', 'rr', 'ff')
    attrs = tuple(filter(lambda x: not x.startswith('_'), __slots__))
    _getter = attrgetter(*attrs)

    def values(self):
        return self._getter(self)

    def items(self):
        return dict(zip(self.attrs, self.values()))

    def __init__(self, tlen_far=700):
        self._tlen_far = tlen_far
        self.unmapped = self.jumped = self.far = self.overlap = self.penetrate = self.rr = self.ff = 0

    def add(self, rec):
        if rec.mate_is_unmapped:
            self.unmapped += 1
            return
        if rec.rnext != rec.tid:
            self.jumped += 1
            return

        # check orientation
        if not rec.is_reverse:
            if not rec.mate_is_reverse:
                self.ff += 1
            elif rec.tlen > self._tlen_far:
                self.far += 1
            elif rec.pnext < rec.aend:
                self.overlap += 1
            elif rec.tlen < 0:
                self.penetrate += 1
        else:
            if rec.mate_is_reverse:
                self.rr += 1
            elif - rec.tlen > self._tlen_far:
                self.far += 1
            elif rec.aend - rec.pnext < - 2 * rec.tlen:  # adhoc
                self.overlap += 1
            elif rec.aend < rec.pnext:
                self.penetrate += 1

    def __str__(self):
        return str(self.items())


class BreakPointCounter(ReadCounterBase):
    def __init__(self, start, window):
        self.start = start
        self.end = start + window
        self.lclips = []

    def add(self, rec):
        self.count += 1
        is_lclip = (rec.qstart > 0)
        is_rclip = (rec.qend < rec.rlen)
        is_clip = is_lclip or is_rclip
        is_mapq1 = (rec.mapq <= 1)
        if is_lclip and self.start <= rec.pos < self.end:
            lclips.append()
        self.lclip += int(is_lclip)
        self.rclip += int(is_rclip)
        self.bclip += int(is_lclip and is_rclip)
        self.reverse += rec.is_reverse
        self.mapq1 += int(is_mapq1)
        self.mate.add(rec)
        self.covlen += covlen
        if is_mapq1:
            self.covlen_mapq1 += covlen

    def __str__(self):
        return '\t'.join((str(self.start), str(self.end), str(self.__dict__)))


class ReadCountGenerator(object):
    def __init__(self, sam, rname, start=0, end=None, window=50, mass='middle', skip_flag=0x904, counter_cls=ReadCounter):
        #self._samit = samit
        if end is None:
            rlens = dict(zip(sam.references, sam.lengths))
            end = rlens[rname]
        self._samit = sam.fetch(reference=rname, start=start, end=end)
        self.counters = []
        self.wstart = start if start % window == 0 else start // window * window # 110 -> 100, 100 -> 100, 149 -> 100  # window: 50
        self.wend = end if end % window == 0 else (end // window + 1) * window   # 110 -> 150, 100 -> 100, 149 -> 150  # window: 50
        self.cstart = self.wstart
        self.cend = self.wstart
        self.window = window
        self.skip_count = 0
        self.skip_flag = skip_flag  # unmapped, secondary or supplementary
        get_read_masses = {
            'start': lambda rec: (rec.pos,),
            'end':   lambda rec: (rec.aend - 1,),   # end position of the alignment (note that aend points one past the last aligned residue)
            'middle': lambda rec: ((rec.pos + rec.aend - 1) / 2.,),   # middle point of the alignment
            'overlap': lambda rec: range(rec.pos, rec.aend, window),   # one point per window overlaped for each alignment
        }
        self.get_read_masses = get_read_masses[mass]
        self._bulk_size = 200
        self._counter_cls = counter_cls

    def _flush(self):
        while self.counters:
            yield self._dequeue()

    def _enqueue(self):
        if self.wend is None:
            enque_size = self._bulk_size
        else:
            enque_size = min(self._bulk_size, (self.wend - self.cend) // self.window)
        self.counters.extend([self._counter_cls(self.cend + self.window * i, self.window) for i in xrange(enque_size)])
        self.cend += self.window * enque_size

    def _dequeue(self):
        self.cstart += self.window
        return self.counters.pop(0)

    def _should_skip(self, rec):
        if rec.flag & self.skip_flag:
            self.skip_count += 1
            return True

    def __iter__(self):
        try:
            while 1:
                rec = next(self._samit)
                if self._should_skip(rec):
                    continue

                start = rec.pos  # 0-based
                end = rec.aend
                while self.cend < start:
                    for counter in self._flush():
                        yield counter

                    self._enqueue()
                    #yield self._dequeue()

                while self.cend < end and self.cend < self.wend:
                    self._enqueue()

                while self.counters and self.counters[0].end < start:
                    yield self._dequeue()

                masses = self.get_read_masses(rec)
                for mass in masses:
                    rec_index = int(mass - self.cstart) // self.window
                    if 0 <= rec_index < len(self.counters):
                        self.counters[rec_index].add(rec)

        except StopIteration:
            pass
        except AssertionError as e:
            logging.error('Invalid record was found: (pos: %s, aend: %s)', rec.pos, rec.aend)
            logging.info(rec)
            raise

        for counter in self._flush():
            yield counter



import matplotlib

def init(use='Agg'):
    matplotlib.use(use)


class PdfGrid(object):
    def __init__(self, output, row=4, col=4):
        self.output = output
        self.row = row
        self.col = col
        self._saved = False

    def __enter__(self):
        self.open()
        return iter(self)

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def open(self):
        from matplotlib.backends.backend_pdf import PdfPages
        self._pp = PdfPages(self.output)

    def close(self):
        import matplotlib.pyplot as plt
        if not self._saved:
            self._pp.savefig(plt.gcf())
        self._pp.close()

    def __iter__(self):
        """
        """
        import matplotlib.pyplot as plt
        page = 1
        while 1:
            plt.figure()
            for i in range(self.row * self.col):
                plt.subplot(self.row, self.col, i+1)  # change current axis
                self._saved = False
                yield page, i+1  # offset
            self._pp.savefig(plt.gcf())
            self._saved = True
            page += 1



from __future__ import print_function
from argtools import command, argument
import logging
import os
import tempfile
import pysam
import numpy as np
import pandas as pd
from .. import plot
from .. import sh


class SamMethods(object):
    def __init__(self, sam):
        self._sam = sam
        self.filename = sam.filename
        index1 = '{0}.bai'.format(sam.filename)
        index2 = '{0}.bai'.format(sam.filename.rsplit('.', 2)[0])
        #if os.path.exists(index1):  # no use
        #    self.indexpath = index1
        #elif os.path.exists(index2):
        #    self.indexpath = index2
        #else:
        #    self.indexpath = None

    def _gen_it(self, chroms=None):
        if chroms is None:
            for rec in self._sam:
                yield rec
        else:
            for chrom in chroms:
                sam_it = self._sam.fetch(reference=chrom)
                for rec in sam_it:
                    yield rec

    def estimate_mean_rlen(self, chroms=None, rsample=1000, use_alen=False, skip_flag=0x900):
        """ rough estimattion
        """
        lens = []
        if use_alen:
            get_length = lambda rec: rec.alen
        else:
            get_length = lambda rec: len(rec.seq)

        sam_it = self._gen_it(chroms=chroms)
        try:
            i = rsample
            while i:
                rec = next(sam_it)
                if not (rec.flag & skip_flag):
                    lens.append(get_length(rec))
                    i -= 1

        except StopIteration:
            pass
        self._sam.reset()
        return np.mean(lens)

    def get_reference_length(self, chroms=None):
        if chroms is None:
            return sum(self._sam.lengths) - nlen
        chrom_set = set(chroms)
        tlen = sum(length for ref, length in zip(self._sam.references, self._sam.lengths) if ref in chrom_set)
        return tlen

    def _get_index_stats(self):
        with tempfile.NamedTemporaryFile() as fp:
            logging.info('Emit index_stats to %s', fp.name)
            sh.call(['samtools', 'idxstats', self.filename], stdout=fp.name)
            tab = pd.read_table(fp, header=None, names=['reference', 'length', 'mapped', 'unmapped'])
        return tab

    def get_mapped(self, chroms=None):
        if chroms is None:
            return self._sam.mapped
        stats = self._get_index_stats()
        chrom_set = set(chroms)
        maps = [mapped for chrom, mapped in zip(stats.reference, stats.mapped) if chrom in chrom_set]
        return sum(maps)


class Nbed(object):
    def __init__(self, nbed):
        self._tab = pd.read_table(nbed, header=None, names=['contig', 'start', 'end'])

    def get_length(self, contig):
        return (self._tab.end - self._tab.start)[self._tab.contig == contig].sum()


@command.add_sub
@argument('bamfile')
@argument('-r', '--rsample', type=int, default=1000)
def bam_mean_rlen(args):
    sam = pysam.Samfile(args.bamfile, 'rb')
    samm = SamMethods(sam)
    rlen = samm.estimate_mean_rlen(rsample=args.rsample)
    print (rlen)


@command.add_sub
@argument('bamfile')
def bam_mean_cov(args):
    """ Emit mean coverage (#reads) per base
    """
    sam = pysam.Samfile(args.bamfile, 'rb')
    totlen = sum(sam.lengths)    # reference length (contains N)
    cov = 1. * sam.mapped / totlen # (mapped may contains secondary or supplementary alignments)
    print (cov)


@command.add_sub
@argument('bamfile')
@argument('-r', '--rsample', type=int, default=1000)
@argument('--pacbio', action='store_true', help='use suplementary alignemnt, use only aligned length')
@argument('--nbed', help='N bed regions')
@argument.exclusive(
    argument('--chroms', nargs='+'),
    argument('--autosomal', action='store_true'),
)
def bam_mean_depth(args):
    """ Roughly estimate bam depth

    - estimate mean read length by scanning top {rsample} reads
    - divide the total length by total reference lengths

    # TODO autosomal option
    """
    sam = pysam.Samfile(args.bamfile, 'rb')
    samm = SamMethods(sam)
    if args.autosomal:
        if 'chr1' in sam.references:
            chroms = ['chr' + str(i + 1) for i in range(22)]
        else:
            chroms = [str(i + 1) for i in range(22)]
    else:
        chroms = args.chroms

    if args.pacbio:
        logging.info('PacBio mode')
        rlen = samm.estimate_mean_rlen(rsample=args.rsample, chroms=chroms, use_alen=True, skip_flag=0x104)
    else:
        rlen = samm.estimate_mean_rlen(rsample=args.rsample, chroms=chroms)
    totlen = samm.get_reference_length(chroms=chroms)

    # negating N bases if available
    nlen = 0
    if args.nbed:
        nbed = Nbed(args.nbed)
        if chroms is None:
            nlen = sum(nbed.get_length(ref) for ref in sam.references)
        else:
            chrom_set = set(chroms)
            nlen = sum(nbed.get_length(ref) for ref in sam.references if ref in chrom_set)

    logging.info('Negating N region from total length: %s', nlen)
    mapped = samm.get_mapped(chroms=chroms)
    d = 1. * mapped * rlen / (totlen - nlen) # (mapped may contains secondary or supplementary alignments)
    logging.info('chroms: %s', chroms)
    logging.info('Sampled reads: %s', args.rsample)
    logging.info('Estimated mean read length: %s', rlen)
    logging.info('Total mapped fragments: %s', mapped)
    logging.info('Total refenrece length: %s', totlen)
    print (d)


@command.add_sub
@argument('bamfile')
@argument('-r', '--region')
@argument('-p', '--plot')
def bam_stats_len(args):
    """ Default is length show distribution
    """
    skip_flags = 0x800 | 0x100  # ignore secondary or suppl. alignment
    sam = pysam.Samfile(args.bamfile, mode='rb', check_sq=False)  # check_sq=False is required for PacBio bam
    if args.region:
        samit = sam.fetch(region=args.region)
    else:
        samit = sam

    lengths = []
    for rec in samit:
        logging.debug(rec)
        if rec.flag & skip_flags:
            continue
        length = rec.infer_query_length(always=False)
        if length is None:
            length = rec.query_length
        lengths.append(length)

    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    lens = pd.Series(lengths)
    print (lens.describe())
    if args.plot:
        logging.info('Plot to %s', args.plot)
        plot.init()
        lens.hist(bins=100)
        plt.savefig(args.plot)



from __future__ import print_function
from ..utils import Counter, cached_property
from .fasta import iter_contigs, Contig   # TODO deprecated
from ..bioseq import color_term_dna, mask_same_bases, Fasta, get_aln_pos_text, mask_unknown_seq, SeqLocator, contig2chain, AlignmentAnalyzer
from builtins import zip
from argtools import command, argument
import itertools
import logging
import pandas as pd


@command.add_sub
@argument('fasta', help='all contigs should be the same lengths')
@argument('-n', '--name', default='new_contig', help='name of new contig')
def msa_consensus(args):
    """ Make a consensus sequence from multiple aligned contigs
    """
    contigs = list(iter_contigs(args.fasta))
    seqs = list(contig.seq for contig in contigs)

    def iter_consensus():
        for bases in zip(*seqs):
            counts = Counter(filter(lambda x: x !='-', bases))
            if not counts:  # no bases other than padding '-'
                yield ''
            else:
                cons_base = counts.most_common(1)[0][0]
                yield cons_base

    new_seq = ''.join((base for base in iter_consensus()))

    print ('>' + args.name)
    print (new_seq)


@command.add_sub
@argument('fasta', help='all contigs should be the same lengths')
@argument('-f', '--from-name', default='1', help='name of new contig')
@argument('-t', '--to-name', default='2', help='name of new contig')
def msa2chain(args):
    """ Make chain file from a pair of aligned contigs (default pair is 1st and 2nd)
    """
    with open(args.fasta) as fp:
        fasta = Fasta(fp)
        if args.from_name.isdigit():
            from_contig = fasta.contigs[int(args.from_name) - 1]
        else:
            from_contig = fasta.get(args.from_name)
        if args.to_name.isdigit():
            to_contig = fasta.contigs[int(args.to_name) - 1]
        else:
            to_contig = fasta.get(args.to_name)

        chain = contig2chain(from_contig, to_contig)
        print ('chain', *chain['header'], sep=' ')
        for row in chain['data']:
            print (*row, sep='\t')


# TODO realign option
@command.add_sub
@argument('fasta', help='all contigs should be the same lengths')
@argument('-n', '--names', help='comma separated list of names to calculate')
@argument('-r', '--redundant', action='store_true', help='calculate for all the combinations')
@argument('--ctx-size', type=int, default=3, help='when set this, emit CTX base context sequences for edits')
def msa2edit(args):
    """ Calculate pairwise edit distance from multiple aligned contigs

    Emission records:
        name1 :
        name2 :
        pos  : 123  # pos in alignment
        pos1 : 121  # last end in contig 1 (last pos if seq is deletion)
        pos2 : 122  # last end in contig 2 (last pos if seq is deletion)
        seq1 : A
        seq2 : -
        ctx_size : 3
        ctx1 : AAAAAAA
        ctx2 : AAA-AAA

    # TODO collapsing neighbor variants
    """
    assert args.ctx_size > 0
    with open(args.fasta) as fp:
        fasta = Fasta(fp)
        if args.names:
            names = args.names.split(',')
        else:
            names = fasta.names

        if args.redundant:
            pairs = itertools.product(names, names)
        else:
            pairs = itertools.combinations(names, 2)
        ctx_size = args.ctx_size

        print ('name1', 'name2', 'msa_pos', 'pos1', 'pos2', 'base1', 'base2', 'ctx_size', 'ctx1', 'ctx2', sep='\t')
        for n1, n2 in pairs:
            c1 = fasta.get(n1)
            aln1 = c1.seq.upper()
            c2 = fasta.get(n2)
            aln2 = c2.seq.upper()
            aa = AlignmentAnalyzer(aln1, aln2)
            for pos, b1, b2 in aa.iter_edits():
                ctx = aa.get_context(pos, left=args.ctx_size, right=args.ctx_size+1)
                pos1 = ctx['last_end1']
                pos2 = ctx['last_end2']
                ctx1 = ctx['ctx1']
                ctx2 = ctx['ctx2']
                print (n1, n2, pos, pos1, pos2, b1, b2, ctx_size, ctx1, ctx2, sep='\t')


# TODO realign option
@command.add_sub
@argument('fasta', help='all contigs should be the same lengths')
@argument('-n', '--names', help='comma separated list of names to calculate')
@argument('-r', '--redundant', action='store_true', help='calculate for all the combinations')
def msa2dist(args):
    """ Calculate pairwise edit distance from multiple aligned contigs

        name1 :
        name2 :
        distance : distance
        ins : ins from name1 to name2
        del : del from name1 to name2
        mut : mutation
    """
    with open(args.fasta) as fp:
        fasta = Fasta(fp)
        if args.names:
            names = args.names.split(',')
        else:
            names = fasta.names

        if args.redundant:
            pairs = itertools.product(names, names)
        else:
            pairs = itertools.combinations(names, 2)

        print ('name1', 'name2', 'alen', 'len1', 'len2', 'distance', 'ins', 'del', 'mut', sep='\t')
        for n1, n2 in pairs:
            c1 = fasta.get(n1)
            aln1 = c1.seq.upper()
            c2 = fasta.get(n2)
            aln2 = c2.seq.upper()
            aa = AlignmentAnalyzer(aln1, aln2)

            ins = del_ = mut = 0
            for pos, b1, b2 in aa.iter_edits():
                if b1 == b2:
                    logging.warning('not an edit: (%s, %s, %s)', pos, b1, b2)
                elif b1 == '-':
                    ins += 1
                elif b2 == '-':
                    del_ += 1
                elif b1 != b2:
                    mut += 1
            dist = ins + del_ + mut
            alen = len(aln1)
            len1 = len(aln1.replace('-', ''))
            len2 = len(aln2.replace('-', ''))
            print (n1, n2, alen, len1, len2, dist, ins, del_, mut, sep='\t')


class MSAPileup:
    """
    >>> p1 = MSAPileup(['A', 'A', '*'])
    >>> p1.width, p1.is_variant
    (1, False)
    >>> p1.masks
    (False, False, True)

    >>> p2 = MSAPileup(['A', 'T', '*'])
    >>> p2.width, p2.is_variant
    (1, True)
    >>> p2.masks
    (False, False, True)

    >>> MSAPileup.merge(p1, p2).masks
    (False, False, True)

    >>> p3 = MSAPileup(['*', 'T', 'G'])
    >>> p3.width, p3.is_variant
    (1, True)
    >>> p3.masks
    (True, False, False)

    >>> p4 = MSAPileup.merge(p1, p2, p3)
    >>> p4.width
    3
    >>> p4.bases
    ['AA*', 'ATT', '**G']
    >>> p4.masks
    (False, False, False)
    """

    def __init__(self, bases, pos=None, names=None, mask_char='*'):
        self.bases = bases
        self.width = len(bases[0])
        self.base_counts = Counter(self.bases)
        self.mask_char = mask_char
        self.pos = pos
        self.names = names

    @cached_property
    def masks(self):
        mask_base = self.mask_char * self.width
        return tuple(b == mask_base for b in self.bases)

    @cached_property
    def is_variant(self):
        if self.width > 1:
            warnings.warn('Variant determination for pileup with width > 1 may be inaccurate')
        if self.mask_char * self.width in self.base_counts:
            return len(self.base_counts) > 2
        else:
            return len(self.base_counts) > 1

    @staticmethod
    def merge(*pileups):
        head = pileups[0]
        bases = [''.join(bs) for bs in zip(*(p.bases for p in pileups))]
        return MSAPileup(bases, pos=head.pos, names=head.names, mask_char=head.mask_char)


def iter_msa_pileup(fasta, names=None, variant_only=False, merge_variant=False):
    contigs = fasta.contigs
    if names:
        contig_map = dict((c.name, c) for c in contigs)
        contigs = [contig_map[name] for name in names]
    else:
        names = fasta.names

    mask_char = '*'
    #seq_iters = [iter(c.get_clip_mask_seq(mask_char=mask_char), None) for c in contigs]
    seq_iters = [iter(mask_unknown_seq(c.seq.upper, mask_char=mask_char), None) for c in contigs]
    pos = 0  # 0 start
    buffer = []

    while 1:
        bases = map(next, seq_iters)
        if bases[0] is None:
            break

        pileup = MSAPileup(bases, pos, names, mask_char=mask_char)
        if (pileup.is_variant and merge_variant
            and (not buffer or buffer[0].masks == pileup.masks)):  # mask positions are the same
            buffer.append(pileup)
        else:
            if buffer:
                p = MSAColumn.merge(*buffer)
                yield p
                buffer = []
            if not variant_only:
                yield pileup
        pos += 1
    if buffer:
        p = MSAPileup.merge(*buffer)
        yield p


def msa2df(fasta):
    tab = pd.DataFrame()
    for contig in fasta.contigs:
        contig.name
        seq = contig.get_clip_mask_seq()
        tab[contig.name] = pd.Series(list(seq))
    return tab


@command.add_sub
@argument('fasta', help='all contigs should be the same lengths')
@argument('-p', '--show-position', action='store_true')
@argument('-c', '--color', action='store_true')
@argument('-m', '--mask', action='store_true')
@argument('-u', '--mask-unknown-bases', action='store_true')
#@argument('-d', '--diff', action='store_true', help='only show positions with difference')
#@argument('-r', '--range', action='store_true', help='only show positions with difference')
#@argument('-n', '--ref-name', help='default is consensus')
#@argument('-i', '--ref-index', type=int, help='if specified, ref_name is ignored')   # consensus
def msa_view(args):
    """ Make a consensus sequence from multiple aligned contigs
    """
    def decorate(seq, ref=None):
        if args.mask_unknown_bases:
            seq = mask_unknown_seq(seq, mask_char='_')
        if args.mask and ref is not None:
            seq = mask_same_bases(seq, ref)
        if args.color:
            seq = color_term_dna(seq)
        return seq

    with open(args.fasta) as fp:
        fasta = Fasta(fp)

        it = iter(fasta)
        #it = iter_contigs(fp)
        contig = next(it)
        ref = contig.seq.upper()
        pos_text = get_aln_pos_text(ref, count_del=True)
        print (pos_text['number'], sep='\t')
        print (pos_text['indicator'], sep='\t')
        pos_text = get_aln_pos_text(ref, count_del=False)
        print (pos_text['number'], sep='\t')
        print (pos_text['indicator'], sep='\t')
        print (decorate(ref), 0, contig.name_line, sep='\t')
        for idx, contig in enumerate(it, 1):
            seq = contig.seq.upper()
            #pos_text = get_aln_pos_text(seq)
            #print (pos_text['number'], sep='\t')
            #print (pos_text['indicator'], sep='\t')
            #seq = ' ' * contig.lclip + seq[contig.lclip : len(seq) - contig.rclip] + ' ' * contig.rclip
            print (decorate(seq, ref=ref), idx, contig.name_line.replace('\t', ' '), sep='\t')


@command.add_sub
@argument('fasta', help='all contigs should be the same lengths')
@argument('--layout', help='layout file')
@argument('--fig-x', type=float, default=8.3)
@argument('--fig-y', type=float, default=10.7)
@argument('--title', default='MSA layout plot')
@argument('--adjust-right', type=float, default=.8)
def msa_layout_plot(args):
    """
    layout is tsv file
        name:
        show_name:
        type: [gene,exon]
        parent: name
        show_offset:
        show_diff: [01]   # annotate difference
        # show_variant: [01]
        color: e.g. 'red'
    """
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import matplotlib.patches as patches
    matplotlib.rcParams['figure.figsize'] = (args.fig_x, args.fig_y)  # A4
    matplotlib.rcParams['font.size'] = 10.
    matplotlib.rcParams['axes.titlesize'] = 'medium'
    matplotlib.rcParams['axes.labelsize'] = 'medium'
    matplotlib.rcParams['axes.labelpad'] = 8.
    matplotlib.rcParams['xtick.labelsize'] = 'small'
    matplotlib.rcParams['ytick.labelsize'] = 'small'
    matplotlib.rcParams['legend.fontsize'] = 'small'
    tab = pd.read_table(args.layout)
    out = '{0}.pdf'.format(args.layout)
    logging.info('Create %s', out)

    with open(args.fasta) as fp:
        fasta = Fasta(fp)
        name_contigs = dict(zip(fasta.names, fasta.contigs))
        fig = plt.figure()
        #ax = fig.add_subplot(111, aspect='equal')
        ax = plt.gca()
        y_margin = 100
        offset = 100
        height = 50
        y_ids = []
        id_names = {}

        for row in tab.itertuples():
            print (row)
            name = row.name
            if name not in name_contigs:
                logging.warning('%s is not included in MSA', name)
                continue
            parent = row.parent if isinstance(row.parent, int) else 0
            if parent and (parent not in id_names or id_names[parent] not in name_contigs):
                logging.warning('%s is not included in MSA', name)
                continue
            if not parent:
                y_index = len(y_ids)
                id_names[row.id] = name
                y_ids.append(row.id)
            else:
                y_index = y_ids.index(parent)

            color = row.color
            seq = mask_unknown_seq(name_contigs[name].seq)
            loc = SeqLocator(seq)
            end = len(seq)
            y = y_index * y_margin
            for s, e, t in loc.blocks:
                x = s
                width = e - s
                print (x, y, width, height)
                ax.add_patch(patches.Rectangle((x, y), width, height, facecolor=color))
            if not parent:
                ax.text(end + offset, y + y_margin*.5, row.show_name)
        ax.set_aspect('equal')
        ax.set_xlim(0, end + offset)
        ax.set_ylim(y + y_margin , - y_margin)
        ax.get_yaxis().set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.xaxis.set_ticks_position('bottom')
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        #ax.set_frame_on(False)
        ax.set_title(args.title)
        plt.subplots_adjust(left=.05, right=args.adjust_right)
        plt.savefig(out)



#!/usr/bin/env python
from __future__ import absolute_import
import os
import sys
if __package__ == '':   # hack for relative import
     (search_path, __package__) = os.path.split(os.path.split(os.path.abspath(__file__))[0])
     sys.path.insert(0, search_path)
     __import__(__package__)

from argtools import command
from .tools import fasta
from .tools import msa
from .tools import bam
from .tools import bam_profile
from .tools import bam_summary
from .tools import bam_summary2
from .tools import blast
from .tools import vcf
from .tools import assoc
from .tools import feature
from .tools import imgt
from .tools import segdup
#from .tools import bam_aln
from .tools import gc_profile
#from .tools import covfit_gmm
from .tools import covfit_gmm_v2
#from .tools import covfit_pmm
from .tools import igv
from .tools import misc

def main():
  command.run()

if __name__ == '__main__':
    main()



#!/usr/bin/env python

from __future__ import print_function
import re
import logging
from argtools import command, argument
from itertools import groupby, chain
from collections import namedtuple
from operator import attrgetter, itemgetter


@command
@argument('bed', nargs='?', default='/dev/stdin')
@argument('-s', '--bin-size', type=int, default=50)
@argument('-b', '--both-flank-bins', type=int, default=0)
@argument('-l', '--left-flank-bins', type=int, default=None)
@argument('-r', '--right-flank-bins', type=int, default=None)
@argument('--no-validate', dest='validate', action='store_false', default=True)
def binned_smooth(args):
    """ Smoothing each bedg value with an average value of flanking bins

    Currently, using gz file as input is slow. Following is recommended:
    zcat bedg.gz | <this program>
    """
    left = args.left_flank_bins
    right = args.right_flank_bins
    if left is None:
        left = args.both_flank_bins
    if right is None:
        right = args.both_flank_bins

    logging.info('Validation: %s', args.validate)
    logging.info('Bin size: %s bp', args.bin_size)
    logging.info('Left flank bins: %s (%s bp)', left, left * args.bin_size)
    logging.info('Right flank bins: %s (%s bp)', right, right * args.bin_size)

    smoother = BedgBinnedSmooth(bin_size=args.bin_size, left=left, right=right, validate=args.validate)
    it = read_bedg(args.bed)

    try:
        for (c, s, e, sig) in smoother(it):
            print (c, s, e, '{0:0.3f}'.format(sig), sep='\t')
    except AssertionError as e:
        logging.error('%s', e)
        return 1


BedgRec = namedtuple('BedgRecord', ('chrom', 'start', 'end', 'value'))
def read_bedg(file):
    re_skip = re.compile('#|browser|track')

    if file.endswith('.gz'):
        import gzip
        open = gzip.open
    else:
        import __builtin__
        open = __builtin__.open

    fp = open(file)
    try:
        line = next(fp)
        while re_skip.match(line):
            line = next(fp)

        while 1:
            row = line.rstrip().split()
            yield BedgRec(row[0], int(row[1]), int(row[2]), float(row[3]))
            line = next(fp)
    except StopIteration:
        pass
    finally:
        fp.close()


def with_validation(intervals, bin_size):
    """ Confirm chrom ordered, sorted in start position, and equally binned.
    """
    chrom_set_done = set()
    chrom = None
    start = -bin_size

    for (chrom, ivs) in groupby(intervals, itemgetter(0)):  # group by chrom
        assert chrom not in chrom_set_done, '{0} was appeared again!'.format(chrom)
        chrom_set_done.add(chrom)
        start = -bin_size
        for iv in ivs:
            assert start < iv.start, 'Not sorted in start position!'
            assert iv.end - iv.start == bin_size, 'Interval {0} != bin size: {1}'.format(iv.end - iv.start, bin_size)
            assert iv.start % bin_size == 0, 'Start position {0} should be divided by bin size: {1}'.format(iv.start, bin_size)
            start = iv.start
            yield iv


class BedgBinnedSmooth(object):
    """
    >>> data = [
    ...     BedgRec('chr1', 1000, 1050, 1),  # 20
    ...     BedgRec('chr1', 1050, 1100, 1),  # 21
    ...     BedgRec('chr1', 1100, 1150, 1),  # 22
    ...     BedgRec('chr1', 1200, 1250, 1),  # 24
    ...     BedgRec('chr1', 1950, 2000, 1),  # 39
    ...     BedgRec('chr1', 2000, 2050, 1),  # 40
    ...     BedgRec('chr3', 1000, 1050, 1),  # 20
    ...     BedgRec('chr3', 1050, 1100, 1),  # 21
    ...     BedgRec('chr2', 1000, 1050, 1),  # 20
    ...     BedgRec('chr2', 1050, 1100, 1),  # 21
    ... ]

    >>> for (c, s, e, v) in BedgBinnedSmooth(50)(data): print((c, s, e, '{0:0.3f}'.format(v)))
    ('chr1', 1000, 1050, '1.000')
    ('chr1', 1050, 1100, '1.000')
    ('chr1', 1100, 1150, '1.000')
    ('chr1', 1200, 1250, '1.000')
    ('chr1', 1950, 2000, '1.000')
    ('chr1', 2000, 2050, '1.000')
    ('chr3', 1000, 1050, '1.000')
    ('chr3', 1050, 1100, '1.000')
    ('chr2', 1000, 1050, '1.000')
    ('chr2', 1050, 1100, '1.000')

    >>> for (c, s, e, v) in BedgBinnedSmooth(50, left=1, right=2)(data): print((c, s, e, '{0:0.3f}'.format(v)))
    ('chr1', 900, 950, '0.250')
    ('chr1', 950, 1000, '0.500')
    ('chr1', 1000, 1050, '0.750')
    ('chr1', 1050, 1100, '0.750')
    ('chr1', 1100, 1150, '0.750')
    ('chr1', 1150, 1200, '0.500')
    ('chr1', 1200, 1250, '0.250')
    ('chr1', 1250, 1300, '0.250')
    ('chr1', 1850, 1900, '0.250')
    ('chr1', 1900, 1950, '0.500')
    ('chr1', 1950, 2000, '0.500')
    ('chr1', 2000, 2050, '0.500')
    ('chr1', 2050, 2100, '0.250')
    ('chr3', 900, 950, '0.250')
    ('chr3', 950, 1000, '0.500')
    ('chr3', 1000, 1050, '0.500')
    ('chr3', 1050, 1100, '0.500')
    ('chr3', 1100, 1150, '0.250')
    ('chr2', 900, 950, '0.250')
    ('chr2', 950, 1000, '0.500')
    ('chr2', 1000, 1050, '0.500')
    ('chr2', 1050, 1100, '0.500')
    ('chr2', 1100, 1150, '0.250')
    """

    def __init__(self, bin_size=50, left=0, right=0, validate=True):
        assert left >= 0 and right >= 0
        self.bin_size = bin_size
        self.left = left
        self.right = right
        self.bufsize = left + right
        self.denominator = 1. * (1 + left + right)
        self.validate = validate

    def __call__(self, intervals):
        denom = self.denominator
        bs = self.bin_size
        loffset = self.left  * bs
        roffset = self.right * bs
        if self.validate:
            intervals = with_validation(intervals, bs)

        for (chrom, ivs) in groupby(intervals, itemgetter(0)):  # group by chrom
            emit_start = 0
            buff = []
            sig = 0
            last_iv = BedgRec(chrom, float('inf'), float('inf'), 0.0)  # dummy record for avoid emission after ivs consumed

            for iv in chain(ivs, [last_iv]):
                while buff and emit_start + roffset < iv.start:
                    #sig = sum(v for (s, v) in buff)            # for speed, avoid summation
                    yield (chrom, emit_start, emit_start + bs, sig)
                    emit_start += bs
                    if buff[0][0] < emit_start - loffset:
                        sig -= buff[0][1]
                        buff.pop(0)

                val = iv.value / denom
                sig += val
                buff.append((iv.start, val))
                emit_start = max(emit_start, iv.start - roffset)


if __name__ == '__main__':
    exit(command.run())



from __future__ import print_function
import logging
from collections import namedtuple, defaultdict
from itertools import groupby, chain
from operator import attrgetter, itemgetter
from .utils import deprecated


class IntervalMixin(object):
    """
    >>> iv0 = Interval.with_range(100, 200)
    >>> iv1 = Interval('a', 100, 200, None)
    >>> iv2 = Interval('a', 150, 250, None)
    >>> iv3 = Interval('b', 150, 250, None)
    >>> iv4 = Interval('a', 200, 250, None)
    >>> iv5 = Interval('a', 80, 100, None)
    >>> iv6 = Interval('a', 80, 101, None)

    >>> iv0.contains(iv1)
    True
    >>> iv0.intersects(iv2)
    True
    >>> iv0.intersects(iv3)
    True
    >>> iv0.contains(iv3)
    False
    >>> iv0.intersects(iv4)
    False
    >>> iv0.intersects(iv4, adjacent=True)
    True
    >>> iv2.intersects(iv3)
    False
    >>> iv5.intersects(iv0)
    False
    >>> iv5.leftside_of(iv0)
    True
    >>> iv5.rightside_of(iv0)
    False
    >>> iv5.leftside_of(iv0)
    True
    >>> iv6.leftside_of(iv0)
    False
    >>> iv5.rightside_of(iv0)
    False

    >>> iv0.length
    100

    >>> iv0.intersection(iv2)
    Interval(contig=None, start=150, end=200, data=None)
    >>> iv0.intersection(iv4)  # returns nothing

    >>> iv0.intersection(iv4, adjacent=True)
    Interval(contig=None, start=200, end=200, data=None)

    >>> iv0.intersection(iv6)
    Interval(contig=None, start=100, end=101, data=None)
    >>> iv1.move(start=-150, end=150)
    Interval(contig='a', start=-50, end=350, data=None)
    >>> iv1.move(start=-150, end=150).trim()
    Interval(contig='a', start=0, end=350, data=None)
    >>> iv1.move(start=-150, end=150).trim(start=50, end=200)
    Interval(contig='a', start=50, end=200, data=None)
    >>> iv1.iv
    Interval(contig='a', start=100, end=200, data=None)
    """

    @property
    def iv(self):
        return Interval(self.contig, self.start, self.end, None)   # get interval

    @property
    def length(self):
        return self.end - self.start

    def comparable(self, iv):
        return self.contig is None \
                or iv.contig is None \
                or self.contig == iv.contig

    def contains(self, iv):
        return self.comparable(iv) and (self.start <= iv.start and iv.end <= self.end)

    def intersects(self, iv, adjacent=False):
        return self.comparable(iv) and (
                (self.start <= iv.end and iv.start <= self.end)
                if adjacent
                else (self.start < iv.end and iv.start < self.end)
               )

    def intersection(self, iv, adjacent=False):
        if self.intersects(iv, adjacent=adjacent):
            return Interval(None, max(self.start, iv.start), min(self.end, iv.end), None)

    def leftside_of(self, iv):
        return self.comparable(iv) and self.end <= iv.start

    def rightside_of(self, iv):
        return iv.leftside_of(self)

_missing = object()
class Interval(namedtuple('Interval', 'contig,start,end,data'), IntervalMixin):
    __slots__ = ()

    @classmethod
    def with_range(cls, start, end, contig=None, data=None):
        return cls(contig, start, end, data)

    def move(self, start=0, end=0, data=_missing):
        return Interval(self.contig, self.start + start, self.end + end,
                self.data if data is _missing else data)

    def trim(self, start=0, end=float('inf'), data=_missing):
        return Interval(self.contig, max(self.start, start), min(self.end, end),
                self.data if data is _missing else data)

    def replace(self, contig=_missing, start=_missing, end=_missing, data=_missing):
        return Interval(
                self.contig if contig is _missing else contig,
                self.start if start is _missing else start,
                self.end if end is _missing else end,
                self.data if data is _missing else data)


def min_vacant_integer(lis):
    """
    >>> min_vacant_integer([])
    0
    >>> min_vacant_integer([1, 2, 3])
    0
    >>> min_vacant_integer([0, 1, 3])
    2
    >>> min_vacant_integer([0, 1, 2, 3])
    4
    """
    iset = set(lis)
    i = 0
    while 1:
        if i in iset:
            i += 1
            continue
        else:
            return i

# this may be generalized from interval_cover
@deprecated
def get_containers(ivs):
    """ O(n)

    >>> ivs = [
    ...     Interval('a', 100, 250, None),
    ...     Interval('a', 150, 250, None),
    ...     Interval('b', 150, 250, None),
    ...     Interval('a', 200, 250, None),
    ...     Interval('a', 80, 100, None),
    ...     Interval('a', 80, 101, None),
    ... ]
    >>> ivs = get_containers(ivs)
    >>> ivs['a'] == Interval('a', 80, 250, None)
    True
    >>> ivs['b'] == Interval('b', 150, 250, None)
    True
    """
    starts = {}
    ends = {}
    for iv in ivs:
        start = starts.get(iv.contig)
        if start is None or iv.start < start:
            starts[iv.contig] = iv.start
        end = ends.get(iv.contig)
        if end is None or iv.end > end:
            ends[iv.contig] = iv.end

    return dict((c, Interval(c, starts[c], ends[c], None)) for c in starts)


# TODO rename to interval_levels
def interval_pileup(ivs):
    """
    >>> ivs = [
    ...     Interval('a', 100, 250, None),
    ...     Interval('a', 150, 250, None),
    ...     Interval('b', 150, 250, None),
    ...     Interval('a', 200, 250, None),
    ...     Interval('a', 80, 100, None),
    ...     Interval('a', 80, 101, None),
    ... ]
    >>> interval_pileup(ivs)
    [0, 1, 0, 2, 0, 1]
    """
    # maybe O(n^2) algorithm

    iv_levels = []
    def add_level(iv):
        isecs = [lv1 for iv1, lv1 in iv_levels if iv1.intersects(iv)]
        lv = min_vacant_integer(isecs)
        iv_levels.append((iv, lv))
        return lv

    return [add_level(iv) for iv in ivs]


def interval_cluster(ivs, adjacent=False):
    """
    >>> ivs = [
    ...     Interval('a', 80, 100, None),
    ...     Interval('a', 80, 101, None),
    ...     Interval('a', 100, 250, None),
    ...     Interval('a', 150, 250, None),
    ...     Interval('a', 250, 251, None),
    ...     Interval('b', 150, 270, None),
    ...     Interval('b', 160, 200, None),
    ...     Interval('b', 260, 300, None),
    ... ]
    >>> list(cl for cl, iv in interval_cluster(ivs))
    [1, 1, 1, 1, 2, 3, 3, 3]
    >>> list(cl for cl, iv in interval_cluster(ivs, adjacent=True))
    [1, 1, 1, 1, 1, 2, 2, 2]
    """
    cluster_id = 0
    ivs = check_sorted(ivs)
    for contig, ivs in groupby(ivs, attrgetter('contig')):
        cover_iv = None
        for iv in ivs:
            if cover_iv is None or not iv.intersects(cover_iv, adjacent=adjacent):
                cluster_id += 1
                cover_iv = iv
            else:
                cover_iv = interval_cover((cover_iv, iv))
            yield (cluster_id, iv)


def interval_cover(ivs, validate=False):
    """
    >>> ivs = [
    ...     Interval('a', 80, 100, None),
    ...     Interval('a', 80, 101, None),
    ...     Interval('a', 100, 250, None),
    ...     Interval('a', 150, 250, None),
    ...     Interval('a', 250, 251, None),
    ... ]
    >>> interval_cover(ivs)
    Interval(contig='a', start=80, end=251, data=None)
    """
    ivs = tuple(ivs)
    if validate:
        assert len(set(iv.contig for iv in ivs)) == 1
    contig = ivs[0].contig
    start = min(iv.start for iv in ivs)
    end = max(iv.end for iv in ivs)
    return Interval(contig, start, end, None)


# TODO interval tree
class IntervalStore(object):
    """
    >>> store = IntervalStore([
    ...     Interval('a', 80, 100, None),
    ...     Interval('a', 80, 101, None),
    ...     Interval('a', 150, 250, None),
    ...     Interval('b', 250, 251, None),
    ... ])
    >>> coverings = store.get_coverings(Interval('a', 90, 200, None))
    >>> coverings[0].replace(data=None)
    Interval(contig='a', start=90, end=101, data=None)
    >>> coverings[0].data
    [Interval(contig='a', start=80, end=100, data=None), Interval(contig='a', start=80, end=101, data=None)]
    >>> coverings[1].replace(data=None)
    Interval(contig='a', start=150, end=200, data=None)
    >>> coverings[1].data
    [Interval(contig='a', start=150, end=250, data=None)]
    """

    def __init__(self, ivs):
        # just a list structure
        # maybe interval_tree in the future
        self._ivs = ivs if isinstance(ivs, list) else list(ivs)

    def iter_intersections(self, iv, adjacent=False):
        for iv1 in self._ivs:  # TODO this will be O(n)
            if iv.intersects(iv1, adjacent=adjacent):
                yield iv1

    def get_coverings(self, target_iv):
        coverings = []
        isecs = list(self.iter_intersections(target_iv))
        if not isecs:
            return coverings
        isecs = sorted(isecs, key=lambda x: (x.contig, x.start))
        iv_cls = interval_cluster(isecs)

        for cluster_id, cluster_ivs in groupby(iv_cls, lambda x: x[0]):
            ivs = list(iv for _, iv in cluster_ivs)
            iv_cover = interval_cover(ivs)
            trimmed = iv_cover.trim(start=target_iv.start, end=target_iv.end, data=ivs)
            coverings.append(trimmed)
        return coverings


# TODO set contig order?
def check_sorted(ivs, check_end=False):
    """ Confirm chrom ordered, sorted in start position, and equally binned.
    """
    contig_done = set()

    for contig, ivs in groupby(ivs, attrgetter('contig')):
        assert contig not in contig_done, 'contig: {0} is appeared again!'.format(contig)
        contig_done.add(contig)
        prev_start = -float('inf')
        prev_end = -float('inf')
        for iv in ivs:
            assert prev_start <= iv.start, 'Not sorted in start position!'
            if check_end:
                assert prev_end <= iv.end, 'Not sorted in end position!'
            prev_start = iv.start
            prev_end = iv.end
            yield iv


class ChromCompare:
    def __init__(self, chrom_list):
        self._order = {}
        for i, name in enumerate(chrom_list):
            self._order[name] = i

    def get_order(self, chrom):
        return self._order[chrom]


def parse_bed(bedfile, offset=0):
    """
    Yields: Interval
    """
    with open(bedfile) as fp:
        for line in fp:
            if line.startswith('#'):
                continue
            row = line.rstrip('\r\n').split('\t')
            yield Interval(row[0], int(row[1]) - offset, int(row[2]), row)


def iter_bin_interval_flanks(intervals, bin_size=50, lflank=0, rflank=None, validate=True):
    """
    >>> ivs = [
    ...     Interval('chr1', 1000, 1050, 1),
    ...     Interval('chr1', 1050, 1100, 2),
    ...     Interval('chr1', 1150, 1200, 4),
    ...     Interval('chr1', 1200, 1250, 5),
    ...     Interval('chr1', 1250, 1300, 6),
    ...     Interval('chr1', 1300, 1350, 7),
    ...     Interval('chr1', 1400, 1450, 9),
    ...     Interval('chr1', 1450, 1500, 10),
    ...     Interval('chr2', 1000, 1050, 1),
    ... ]

    >>> for rec in iter_bin_interval_flanks(ivs, bin_size=50):
    ...     print (rec.iv.contig, rec.iv.start, rec.iv.data, tuple(iv.data for iv in rec.flanks), tuple(iv.data for iv in rec.adds), tuple(iv.data for iv in rec.removes))
    chr1 1000 1 (1,) (1,) ()
    chr1 1050 2 (2,) (2,) (1,)
    chr1 1150 4 (4,) (4,) (2,)
    chr1 1200 5 (5,) (5,) (4,)
    chr1 1250 6 (6,) (6,) (5,)
    chr1 1300 7 (7,) (7,) (6,)
    chr1 1400 9 (9,) (9,) (7,)
    chr1 1450 10 (10,) (10,) (9,)
    chr2 1000 1 (1,) (1,) ()

    >>> for rec in iter_bin_interval_flanks(ivs, bin_size=50, lflank=200, rflank=200):
    ...     print (rec.iv.contig, rec.iv.start, rec.iv.data, tuple(iv.data for iv in rec.flanks), tuple(iv.data for iv in rec.adds), tuple(iv.data for iv in rec.removes))
    chr1 1000 1 (1, 2, 4) (1, 2, 4) ()
    chr1 1050 2 (1, 2, 4, 5) (5,) ()
    chr1 1150 4 (1, 2, 4, 5, 6, 7) (6, 7) ()
    chr1 1200 5 (1, 2, 4, 5, 6, 7) () ()
    chr1 1250 6 (2, 4, 5, 6, 7, 9) (9,) (1,)
    chr1 1300 7 (4, 5, 6, 7, 9, 10) (10,) (2,)
    chr1 1400 9 (5, 6, 7, 9, 10) () (4,)
    chr1 1450 10 (6, 7, 9, 10) () (5,)
    chr2 1000 1 (1,) (1,) ()
    """

    if rflank is None:
        rflank = bin_size
    assert lflank >= 0 and rflank >= 0
    assert lflank % bin_size == 0, 'The size of the left flank should be a multiple of the bin size'
    assert rflank % bin_size == 0, 'The size of the right flank should be a multiple of the bin size'

    Record = namedtuple('Record', 'iv flanks adds removes')

    for (contig, ivs) in groupby(intervals, itemgetter(0)):  # group by contig
        emit_index = 0
        flanks = []
        last_iv = Interval(contig=contig, start=float('inf'), end=float('inf'), data=None)  # dummy record for avoid emission after ivs consumed
        adds = []
        removes = []

        for iv in chain(ivs, [last_iv]):
            while emit_index < len(flanks) and flanks[emit_index].start + rflank < iv.end:
                while flanks and flanks[0].start < flanks[emit_index].start - lflank:
                    removes.append(flanks.pop(0))
                    emit_index -= 1

                logging.debug('%s', (flanks[emit_index].start, emit_index, iv.start, flanks, adds, removes))
                if emit_index >= len(flanks):
                    break
                yield Record(iv=flanks[emit_index], flanks=tuple(flanks), adds=tuple(adds), removes=tuple(removes))
                emit_index += 1
                adds = []
                removes = []

            if iv is not last_iv:
                adds.append(iv)
                flanks.append(iv)


def align_intervals(ivs1, ivs2, chrom_cmp):
    """
    >>> ivs1 = [
    ...     Interval('chr1', 1000, 1050, '1_1'),
    ...     Interval('chr1', 1000, 1050, '1_1b'),
    ...     Interval('chr1', 1050, 1100, '1_2'),
    ...     Interval('chr1', 1150, 1200, '1_4'),
    ...     Interval('chr1', 1150, 1200, '1_4b'),
    ...     Interval('chr1', 1200, 1250, '1_5'),
    ...     Interval('chr1', 1250, 1300, '1_6'),
    ...     Interval('chr1', 1300, 1350, '1_7'),
    ...     Interval('chr1', 1400, 1450, '1_9'),
    ...     Interval('chr1', 1450, 1500, '1_10'),
    ...     Interval('chr2', 1000, 1050, '2_1'),
    ...     Interval('chr2', 1050, 1100, '2_2'),
    ...     Interval('chr2', 1100, 1150, '2_3'),
    ...     Interval('chr3', 1000, 1050, '3_1'),
    ... ]
    >>> ivs2 = [
    ...     Interval('chr1', 1050, 1100, '1_2'),
    ...     Interval('chr1', 1150, 1200, '1_4'),
    ...     Interval('chr1', 1200, 1250, '1_5'),
    ...     Interval('chr1', 1250, 1300, '1_6'),
    ...     Interval('chr1', 1300, 1350, '1_7'),
    ...     Interval('chr1', 1350, 1400, '1_8'),
    ...     Interval('chr1', 1400, 1450, '1_9'),
    ...     Interval('chr1', 1450, 1500, '1_10'),
    ...     Interval('chr1', 1500, 1550, '1_11'),
    ...     Interval('chr1', 1550, 1600, '1_12'),
    ...     Interval('chr3', 1000, 1050, '3_1'),
    ...     Interval('chr3', 1050, 1100, '3_2'),
    ... ]
    >>> l = list(align_intervals(ivs1, ivs2, ChromCompare(['chr1', 'chr2', 'chr3'])))
    >>> for iv in l:
    ...     print (iv.contig, iv.start, iv.end, iv.data[0], iv.data[1])
    chr1 1000 1050 1_1 None
    chr1 1000 1050 1_1b None
    chr1 1050 1100 1_2 1_2
    chr1 1150 1200 1_4 1_4
    chr1 1150 1200 1_4b None
    chr1 1200 1250 1_5 1_5
    chr1 1250 1300 1_6 1_6
    chr1 1300 1350 1_7 1_7
    chr1 1350 1400 None 1_8
    chr1 1400 1450 1_9 1_9
    chr1 1450 1500 1_10 1_10
    chr1 1500 1550 None 1_11
    chr1 1550 1600 None 1_12
    chr2 1000 1050 2_1 None
    chr2 1050 1100 2_2 None
    chr2 1100 1150 2_3 None
    chr3 1000 1050 3_1 3_1
    chr3 1050 1100 None 3_2
    """
    ivs1 = groupby(iter(ivs1), lambda x: x[0])
    ivs2 = groupby(iter(ivs2), lambda x: x[0])
    chrom1, iivs1 = next(ivs1, (None, None))
    chrom2, iivs2 = next(ivs2, (None, None))

    while 1:
        i1 = chrom_cmp.get_order(chrom1) if chrom1 is not None else -1
        i2 = chrom_cmp.get_order(chrom2) if chrom2 is not None else -1

        if i1 < i2:
            for iv1 in iivs1:
                yield Interval(chrom1, iv1.start, iv1.end, data=(iv1.data, None))
            chrom1, iivs1 = next(ivs1, (None, None))
            continue
        elif i2 < i1:
            for iv2 in iivs2:
                yield Interval(chrom2, iv2.start, iv2.end, data=(None, iv2.data))
            chrom2, iivs2 = next(ivs2, (None, None))
            continue
        elif i1 == i2 == -1:
            return

        iv1 = next(iivs1, None)
        iv2 = next(iivs2, None)
        while 1:
            if iv1 is None and iv2 is None:
                break
            elif iv1 is None:
                for iv in chain((iv2, ), iivs2):
                    yield Interval(iv.contig, iv.start, iv.end, data=(None, iv.data))
                break
            elif iv2 is None:
                for iv in chain((iv1, ), iivs1):
                    yield Interval(iv.contig, iv.start, iv.end, data=(iv.data, None))
                break

            o1 = (iv1.start, iv1.end)
            o2 = (iv2.start, iv2.end)
            if o1 == o2:
                yield Interval(chrom1, iv1.start, iv1.end, data=(iv1.data, iv2.data))
                iv1 = next(iivs1, None)
                iv2 = next(iivs2, None)
                continue
            elif o1 < o2:
                yield Interval(chrom1, iv1.start, iv1.end, data=(iv1.data, None))
                iv1 = next(iivs1, None)
                continue
            else:
                yield Interval(chrom2, iv2.start, iv2.end, data=(None, iv2.data))
                iv2 = next(iivs2, None)
                continue
        chrom1, iivs1 = next(ivs1, (None, None))
        chrom2, iivs2 = next(ivs2, (None, None))



from __future__ import print_function, absolute_import
import socket
import os
import sys
import signal
import logging
from .utils import get_empty_port


class IGVServer(object):
    def __init__(self, igv_cmd='igv.sh', port=None, outdir=None):
        import subprocess
        from threading import Thread
        import time

        if port is None:
            port = get_empty_port(default=60151)

        self.port = port
        def readit(ffrom, fto, self):
            for line in iter(ffrom.readline, ''):
                if "Listening on port" in line:
                    self._wait = False
                if line.rstrip():
                    fto.write(line)
            ffrom.close()

        if self.port:
            igv_cmd = ' '.join((igv_cmd, '--port', str(self.port)))
        logging.info('igv_cmd: %s', igv_cmd)
        self._proc = p = subprocess.Popen(igv_cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        self._wait = True
        self._tout = tout = Thread(target=readit, args=(p.stdout, sys.stderr, self))
        self._terr = terr = Thread(target=readit, args=(p.stderr, sys.stderr, self))
        #tout.daemon = terr.daemon = True
        tout.start()
        terr.start()
        while p.poll() is None and self._wait:
            time.sleep(10)
            logging.info("waiting...")
        logging.info('port: %s', self.port)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self._proc.send_signal(signal.SIGINT)
        while self._tout.is_alive():
            self._tout.join(1)
        while self._terr.is_alive():
            self._terr.join(1)

    def get_client(self):
        return IGVClient(port=self.port)


class IGVClient(object):
    r"""
    Simple wrapper to the IGV (http://www.broadinstitute.org/software/igv/home)
    socket interface (http://www.broadinstitute.org/software/igv/PortCommands)
    Successful commands return 'OK'
    """
    _socket = None
    _outdir = None

    def __init__(self, host=None, port=None, outdir=None):
        self.host = host or '127.0.0.1'
        self.port = port or 60151
        self.commands = []
        self._connect()
        if outdir is None:
            outdir = os.path.abspath(os.curdir)
        if outdir:
            self._set_outdir(outdir)

    def _connect(self):
        if self._socket:
            self._socket.close()
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((self.host, self.port))

    def close(self):
        self._socket.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _set_outdir(self, outdir):
        outdir = os.path.abspath(outdir)
        if not os.path.exists(outdir):
            os.makedirs(outdir)
            logging.info('Created %s', outdir)

        self.send('snapshotDirectory %s' % outdir)
        self._outdir = outdir

    def send(self, cmd):
        logging.info('cmd:', cmd)
        self.commands.append(cmd)
        self._socket.send(cmd + '\n')
        return self._socket.recv(4096).rstrip('\n')

    def save(self, path=None):
        """ .png, .jpg, or .svg
        """
        if path is not None:
            # igv assumes the path is just a single filename, but
            # we can set the snapshot dir. then just use the filename.
            dirname = os.path.dirname(path)
            dirname = os.path.abspath(dirname)
            org_dir = self._outdir
            if dirname != org_dir:
                self._set_outdir(dirname)
            status = self.send('snapshot ' + os.path.basename(path))
            if org_dir is not None and org_dir != dirname:
                self._set_outdir(org_dir)
            return status
        else:
            return self.send('snapshot')

    snapshot = save

    def new(self):
        return self.send('new')

    def genome(self, name):
        return self.send('genome {0}'.format(name))

    def load(self, url):
        return self.send('load {0}'.format(url))

    def goto(self, *loci):
        if not loci:
            loci = ['all']
        return self.send('goto {0}'.format(' '.join(map(str, loci))))

    def set_height(self, height=1000):
        return self.send('maxPanelHeight {0}'.format(height))

    def sort(self, option='base', locus=''):
        """
        options is one of: base, position, strand, quality, sample, and
        readGroup.
        base, position, strand, quality, sample, readGroup, AMPLIFICATION, DELETION, EXPRESSION, SCORE, MUTATION_COUNT
        """
        assert option in ("base", "position", "strand", "quality", "sample",
                         "readGroup")
        return self.send('sort {0} {1}'.format(option, locus))

    def expand(self, track=''):
        self.send('expand {0}'.format(track))

    def collapse(self, track=''):
        self.send('collapse {0}'.format(track))

    def squish(self, track=''):
        self.send('squish {0}'.format(track))

    def view_as_pairs(self, track=''):
        self.send('viewaspairs {0}'.format(track))

    def mark_region(self, chrom, start, end):
        self.send('region {0} {1} {2}'.format(chrom, start, end))



import logging
from argtools import command, argument
from ..blast import BlastTabFile
from ..utils import iter_tabs
import pysam
from galgo import plot
from ..samutil import parse_region
import numpy as np
import io


def _read_genomes(fai_file):
    with io.open(fai_file) as reader:
        return [(rec[0], int(rec[1])) for rec in iter_tabs(reader)]


@command.add_sub
@argument('blast_tab')
@argument('-g', '--genome', required=True, help='genome file (.fai file)')
@argument('-o', '--outfile', default='/dev/stdout')
@argument('--max-evalue', type=float)
def blast2sam(args):
    header = {
            'HD': {'VN': '1.0'},
    }

    max_evalue = args.max_evalue
    def test(rec):
        if max_evalue is None:
            return True
        elif dict(rec.tags)['ZE'] <= max_evalue:
            return True

    with io.open(args.blast_tab) as reader:
        tab = BlastTabFile(reader)
        genomes = _read_genomes(args.genome)
        SQ = [{'LN': length, 'SN': ref} for ref, length in genomes]
        header['SQ'] = SQ
        reference_ids = dict((ref, tid) for tid, (ref, length) in enumerate(genomes))

        with pysam.AlignmentFile(args.outfile, 'wh', header=header) as out:
            for rec in tab.iter_sam_record(reference_ids):
                if test(rec):
                    out.write(rec)


@command.add_sub
@argument('blast_tabs', nargs='+', help='blastn with -outfmt 7')
@argument('--max-evalue', type=float)
@argument('--max-overhang-ratio', type=float)
@argument('--min-qratio', type=float)
def blast_table(args):
    max_evalue = args.max_evalue
    import sys
    for i, blast_tab in enumerate(args.blast_tabs):
        with io.open(blast_tab) as reader:
            try:
                tab = BlastTabFile(reader)
                for j, (qname, qtab) in enumerate(tab.iter_query_tables()):
                    header = i == 0 and j == 0
                    mode = 'w' if i == 0 and j == 0 else 'wa'
                    if args.max_evalue is not None:
                        qtab = qtab[qtab.evalue <= args.max_evalue]
                    if args.max_overhang_ratio is not None:
                        qtab = qtab[(qtab.left_overhang + qtab.right_overhang) / qtab.qlen <= args.max_overhang_ratio]
                    if args.min_qratio is not None:
                        qtab = qtab[qtab.qratio >= args.min_qratio]
                    qtab.to_csv(sys.stdout, header=header, index=False, mode=mode, sep='\t')
            except Exception as e:
                logging.warning('Error for %s (%s:%s)', blast_tab, type(e), e)


class Measure():
    def __init__(self, a, b):
        self._a = a
        self._b = b
        self._l = b - a

    def __call__(self, value):
        return 1. * (value - self._a) / self._l


# TODO load gtf file for annotation
# TODO add gridline (and ticks for high identity features)
@command.add_sub
@argument('blast_tab', help='blastn with -outfmt 7')
def blast2dotplot(args):
    import sys
    plot.init()
    import matplotlib.pyplot as plt
    from matplotlib.lines import Line2D
    output = '{0}.pdf'.format(args.blast_tab)
    match_rate = Measure(.8, 1.)
    with io.open(args.blast_tab) as reader:
        tab = BlastTabFile(reader)
        for j, (qname, qtab) in enumerate(tab.iter_query_tables()):
            qreg, qstart, qend = parse_region(qname)
            sreg, sstart, send = parse_region(qtab.sname[0])  # FIXME for multiple case
            qoffset = qstart or 0
            soffset = sstart or 0
            plt.figure(figsize=(8., 8.))
            ax = plt.gca()
            ax.set_xlim(sstart, send)  # FIXME for start, end were unavailable
            ax.set_ylim(qstart, qend)  # FIXME for start, end were unavailable
            for row in qtab.itertuples():
                linestyle = '-'
                lw = 2.
                color = ('r' if row.is_reverse else 'b')
                alpha = match_rate(1. - 1. * row.edit_distance / row.aln_len)
                alpha = min(max(0, alpha), 1)
                logging.info('Plotting %s (alpha: %s)', ((row.sstart, row.send), (row.qstart, row.qend)), alpha)
                ax.add_line(Line2D((soffset + row.sstart, soffset + row.send), (qoffset + row.qstart, qoffset + row.qend),
                    linestyle=linestyle, color=color, linewidth=lw, alpha=alpha))
            plt.savefig(output)



from __future__ import print_function
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
try:
    from collections import Counter
except ImportError:
    from counter import Counter
try:
    from UserDict import DictMixin
except ImportError:
    from collections import MutableMapping as DictMixin
from builtins import zip
from six.moves import zip_longest
from itertools import chain
import os
import re
import tempfile
import logging
import warnings
import numpy as np
import functools
from .sselogsumexp import logsumexp

_term_colors = {
    'clear': '\033[0m',
    'black': '\033[30m',
    'red': '\033[31m',
    'green': '\033[32m',
    'yellow': '\033[33m',
    'blue': '\033[34m',
    'purple': '\033[35m',
    'cyan': '\033[36m',
    'white': '\033[37m'
}


class attrdict(object, DictMixin):
    """ Dictionary class whose properties are accessible as attribute

    #TODO implement merge method if needed

    >>> d0 = attrdict()  # empty
    >>> d0
    {}

    >>> d1 = attrdict([('a', 1), ('b', {'key': 'value'}), ('c', 3)], __recursive__=True)  # init with (key, value) pairs
    >>> d1.a, d1.b, d1.b.key, d1.c                                    # access with attribute
    (1, {'key': 'value'}, 'value', 3)
    >>> d1['a']
    1

    >>> d2 = attrdict({'a': 1, 'b': 2, 'd': ['e', {'f': 2}]}, __recursive__=True)         # init with dict
    >>> d3 = attrdict(a=1, b=2, d=['e', {'f': 2}], __recursive__=True)                    # init with kwds style
    >>> d2 == d3
    True
    >>> d3.d[1]
    {'f': 2}

    >>> d1.update(d2)   # override dict
    >>> d1.a, d1.b, d1.c, d1.d
    (1, 2, 3, ['e', {'f': 2}])
    """
    def __init__(self, *args, **kwds):
        if args:
            self.__dict__ = dict(args[0])
        if kwds:
            self.__dict__.update(**kwds)

        recursive = kwds.pop('__recursive__', False)
        if recursive:
            for (k, v) in self.items():
                if isinstance(v, dict):
                    self[k] = attrdict(v)
                elif isinstance(v, (list, tuple, set, frozenset)):
                    self[k] = v.__class__((attrdict(el) if isinstance(el, dict) else el) for el in v)

    def __iter__(self):
        return iter(self.__dict__)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def __repr__(self):
        return repr(dict(self.items()))


class defaultattrdict(attrdict):
    """ Dictionary class whose properties are accessible as attribute

    #TODO implement merge method if needed

    >>> d0 = defaultattrdict(dict)  # empty
    >>> d0
    {}

    >>> d1 = defaultattrdict(list)
    >>> d1.a.append(1)
    >>> d1.a.append(2)
    >>> d1['a']
    [1, 2]
    >>> d1.a
    [1, 2]
    >>> d1.b
    []
    """
    def __init__(self, default):
        self._default = default
        super(defaultattrdict, self).__init__()

    def __getitem__(self, key):
        return getattr(self, key)

    def __getattr__(self, key):
        if key.startswith('_'):
            return self.__dict__[key]
        try:
            return super(defaultattrdict, self).__getattr__(key)
        except AttributeError:
            setattr(self, key, self._default())
            return self[key]

    def __iter__(self):
        return (k for k in iter(self.__dict__) if not k.startswith('_'))


def color_term(term, color):
    return ''.join((_term_colors[color], term, _term_colors['clear']))


def fill_text(text, length, char=' '):
    """
    >>> fill_text('abc', 5, char='+')
    'abc++'
    >>> fill_text('abc', 2, char='+')
    'abc'
    """
    return (text + char * length)[:max(len(text), length)]


def remove_suffix(text, suffix):
    """
    >>> remove_suffix('NA12878.sorted.bam', '.bam')
    'NA12878.sorted'
    >>> remove_suffix('NA12878.sorted.bam', '.sam')
    'NA12878.sorted.bam'
    >>> remove_suffix('NA12878.sorted.bam', '')
    'NA12878.sorted.bam'
    """
    if suffix and text.endswith(suffix):
        return text[:-len(suffix)]
    else:
        return text


def iter_tabs(it, sep='\t', skip_if=None):
    if skip_if is None:
        skip_if = lambda line : False

    for line in it:
        if skip_if(line):
            continue
        yield line.rstrip('\r\n').split(sep)

def with_header(rows, container=OrderedDict, use_header=False):
    """
    >>> tab = iter_tabs(['a,b,c', '1,2,3', '4,5,6,7'], sep=',')
    >>> recs = list(with_header(tab))
    >>> recs[0]['a']
    '1'
    >>> list(recs[1].values())
    ['4', '5', '6']
    """
    rows = iter(rows)
    keys = next(rows)
    if use_header:
        yield tuple(keys)
    for vals in rows:
        yield container(zip(keys, vals))


def _safetype(type):
    def f(v, default=None):
        try:
            return type(v)
        except Exception as e:
            return default
    return f

safeint = _safetype(int)
safefloat = _safetype(float)

def safediv(a, b, error=float('nan')):
    try:
        return a / b
    except ZeroDivisionError:
        return error


def weighted_choice(weights, size=1):
    """ weights is an instance of numpy.array
    >>> np.random.seed(0)
    >>> dict(Counter(weighted_choice(np.array([1, 0, 0]), 10000)))
    {0: 10000}
    >>> np.random.seed(0)
    >>> dict(Counter(weighted_choice(np.array([1.5, 0, 3.0, 0.5]), 10000)))
    {0: 3055, 2: 5943, 3: 1002}
    """
    t = weights.cumsum()
    s = weights.sum()
    return np.searchsorted(t, np.random.rand(size) * s)


_missing = object()
def chunked(seq, size, fillvalue=_missing):
    """
    >>> [tuple(lis) for lis in chunked([1, 2, 3, 4, 5, 6, 7], 3)]
    [(1, 2, 3), (4, 5, 6)]
    >>> [tuple(lis) for lis in chunked([1, 2, 3, 4, 5, 6, 7], 3, fillvalue=None)]
    [(1, 2, 3), (4, 5, 6), (7, None, None)]
    """
    if fillvalue is _missing:
        return zip(*([iter(seq)] * size))
    else:
        return zip_longest(fillvalue=fillvalue, *([iter(seq)] * size))


def bucket(it, size):
    """
    >>> [tuple(lis) for lis in bucket([1, 2, 3, 4, 5, 6, 7], 3)]
    [(1, 2, 3), (4, 5, 6), (7,)]
    """
    assert size > 0
    i = 0
    stock = []
    for el in it:
        stock.append(el)
        i += 1
        if i == size:
            yield stock
            stock = []
            i = 0
    if stock:
        yield stock


def unzip(it, size):
    """
    >>> a, b = unzip(zip((1, 2, 3, 4, 5), ('a', 'b', 'c')), 2)
    >>> a
    [1, 2, 3]
    >>> b
    ['a', 'b', 'c']
    """
    els = tuple([] for _ in xrange(size))
    for el in it:
        for i in xrange(size):
            els[i].append(el[i])
    return els


def iter_pairs(seq, last=None):
    """
    >>> list(iter_pairs((1, 2, 3, 4, 5)))
    [(1, 2), (2, 3), (3, 4), (4, 5), (5, None)]
    >>> list(iter_pairs((1,)))
    [(1, None)]
    """
    it = iter(seq)
    prev = next(it)
    for el in it:
        yield (prev, el)
        prev = el
    yield (prev, last)


def iter_counts(seq, same=lambda a, b: a==b):
    """
    >>> list(iter_counts(('a', 'a', 'b', 'A', 'a', None, 'c')))
    [('a', 2), ('b', 1), ('A', 1), ('a', 1), (None, 1), ('c', 1)]
    >>> list(iter_counts(('a', 'a', 'b', 'A', 'a', 'c'), same=lambda a, b: a.upper()==b.upper()))
    [('a', 2), ('b', 1), ('A', 2), ('c', 1)]
    """
    it = iter(seq)
    try:
        prev = next(it)
    except StopIteration:
        return

    count = 1
    for el in it:
        if same(prev, el):
            count += 1
        else:
            yield (prev, count)
            prev = el
            count = 1
    yield (prev, count)


def iter_uniques(iterable):
    """ Creates an ordered set from a list of tuples or other hashable items

    >>> ''.join(iter_uniques('abracadabra'))
    'abrcd'
    >>> ''.join(iter_uniques('simsalabim'))
    'simalb'
    """
    seen = set()

    for item in iterable:
        #Save unique items in input order
        if item not in seen:
            seen.add(item)
            yield item


def collect_while(pred, it):
    """
    >>> cols, it = collect_while(lambda x: x=='A', 'AAABAAAA')
    >>> ''.join(cols)
    'AAA'
    >>> next(it)
    'B'
    >>> cols, it = collect_while(lambda x: x=='A', it)
    >>> ''.join(cols)
    'AAAA'
    >>> next(it, 'end')
    'end'
    """
    it = iter(it)
    lis = []
    try:
        while 1:
            el = next(it)
            if not pred(el):
                break
            lis.append(el)
        return lis, chain((el,), it)
    except StopIteration as e:
        if lis:
            return lis, it
        else:
            raise e

def skip_until(pred, it):
    """
    >>> el, it = skip_until(lambda x: x=='B', 'AAABAAAA')
    >>> el
    'B'
    >>> el, it = skip_until(lambda x: x=='A', it)
    >>> el
    'A'
    >>> ''.join(it)
    'AAA'
    """
    cols, it = collect_while(make_not(pred), it)
    el = next(it)
    return el, it

_blank_pat1 = re.compile('^\s*$')
_blank_pat2 = re.compile('\s+')

def isblank(line):
    '''
    >>> all(map(isblank, [' ', '', '  ', '\\n']))
    True
    >>> any(map(isblank, [' A ', 'A', ' ABC', 'ABCD \\n']))
    False
    '''
    return bool(_blank_pat1.match(line))

def make_not(fn):
    """ Invert return value of input function
    >>> notblank = make_not(isblank)
    >>> any(map(notblank, [' ', '', '  ', '\\n']))
    False
    >>> all(map(notblank, [' A ', 'A', ' ABC', 'ABCD \\n']))
    True
    """
    return lambda *args, **kwds: not fn(*args, **kwds)

def blank_split(string, maxsplit=0):
    """ 
    >>> blank_split('A   B  C')
    ['A', 'B', 'C']
    >>> blank_split('   A  B C   ')
    ['A', 'B', 'C']
    """
    return _blank_pat2.split(string.strip(), maxsplit=maxsplit)


# decorator
def deprecated(fn):
    def g(*args, **kwds):
        warnings.warn("deprecated", DeprecationWarning)
        return fn(*args, **kwds)
    return g


def memoize(fn, *args):
    """ A simple memoize function
    """
    _d = {}
    @functools.wraps(fn)
    def f(*args):
        _d[args] = fn(*args)
        return _d[args]
    return f


# taken from https://github.com/mitsuhiko/werkzeug
class cached_property(object):
    """A decorator that converts a function into a lazy property.  The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value::

        class Foo(object):

            @cached_property
            def foo(self):
                # calculate something important here
                return 42

    The class has to have a `__dict__` in order for this property to
    work.
    """ 

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


def get_empty_port(default=None, port_range=None):
    """ get port
    """
    import psutil
    conns = psutil.net_connections()
    used_ports = set([conn.laddr[1] for conn in conns])
    if default and default not in used_ports:
        return default
    if port_range is None:
        port_range = range(49152, 65535+1)  # ephemeral ports
    for port in port_range:
        if port not in used_ports:
            return port


def file_split(input, nlines=1000, prefix='', suffix='', dir=None, header=None):
    """
    >>> data = range(50)
    >>> d = tempfile.mkdtemp()
    >>> results = list(file_split(data, nlines=25, prefix='abc.', suffix='.num', dir=d))
    >>> [os.path.basename(f) for f in results]
    ['abc.1.num', 'abc.2.num']
    >>> open(results[0]).read() == ''.join(map(str, range(25)))
    True
    >>> open(results[1]).read() == ''.join(map(str, range(25, 50)))
    True
    >>> os.remove(results[0])
    >>> os.remove(results[1])
    >>> os.removedirs(d)
    """
    assert nlines > 0
    if dir is None:
        dir = tempfile.mkdtemp()
    logging.info('Write split files in {0}'.format(dir))

    def get_next(n):
        name = os.path.join(dir, '{0}{1}{2}'.format(prefix, n, suffix))
        writer = open(name, 'w+')
        if header is not None:
            print (header, file=writer)
        return name, writer

    name = writer = None
    try:
        n = 1
        l = 0
        name, writer = get_next(n)
        for line in input:
            if l == nlines:
                writer.close()
                yield name
                n += 1
                name, writer = get_next(n)
                l = 0
            print (line, end='', file=writer)
            l += 1
        writer.close()
        writer = None
        yield name
    finally:
        if writer is not None:
            writer.close()


def get_process_pool(njobs, init=None):
    from multiprocess import Pool
    import signal
    def init_pool():
        logging.info('Init pool')
        signal.signal(signal.SIGINT, signal.SIG_IGN)
    if init is None:
        init = init_pool
    pool = Pool(njobs, init_pool)
    return pool



#!/usr/bin/env python
from __future__ import print_function
from argtools import command, argument
from itertools import groupby, chain
from collections import namedtuple, defaultdict
import logging
from operator import attrgetter, itemgetter
import re
import sys
import os
import gzip
import numpy as np
from ..utils import iter_tabs, with_header, safeint, safefloat, safediv, weighted_choice, Counter
from ..interval import Interval, iter_bin_interval_flanks, align_intervals, ChromCompare
from pysam import Fastafile, Tabixfile



def validated_bin_intervals(intervals, bin_size):
    """ Confirm chrom ordered, sorted in start position, and equally binned.
    """
    chrom_set_done = set()
    chrom = None
    start = -bin_size

    for chrom, ivs in groupby(intervals, itemgetter(0)):  # group by chrom
        assert chrom not in chrom_set_done, '{0} was appeared again!'.format(chrom)
        chrom_set_done.add(chrom)
        start = -bin_size
        for iv in ivs:
            assert start < iv.start, 'Not sorted in start position!'
            assert iv.end - iv.start == bin_size, 'Interval {0} != bin size: {1}'.format(iv.end - iv.start, bin_size)
            assert iv.start % bin_size == 0, 'Start position {0} should be divided by bin size: {1}'.format(iv.start, bin_size)
            start = iv.start
            yield iv


@command.add_sub
@argument('nuc', help='output of fa_nuc')
@argument('-s', '--bin-size', type=int, default=50)
@argument('-l', '--left-flank', type=int, default=200, help='size of the left side of the window start (a multiple of the window size)')
@argument('-r', '--right-flank', type=int, default=200, help='size or the right side of the window start (a multiple of the window size)')
@argument('--no-validate', dest='validate', action='store_false', default=True)
def add_gc_flanks(args):
    """
    Required columns:

    """
    logging.info('Validation: %s', args.validate)
    logging.info('Window size: %s bp', args.bin_size)
    logging.info('Left flank: %s bp', args.left_flank)
    logging.info('Right flank: %s bp', args.right_flank)

    assert args.left_flank % args.bin_size == 0, 'Size of the left flank should be a multiple of the bin size'
    assert args.right_flank % args.bin_size == 0, 'Size of the right flank should be a multiple of the bin size'

    lbins = args.left_flank / args.bin_size
    rbins = args.right_flank / args.bin_size

    lines = open(args.nuc)
    nucs = with_header(iter_tabs(lines), use_header=True)
    header = next(nucs)

    def iter_ivs(nucs):
        for chrom, nucs in groupby(nucs, lambda x: x['chrom']):
            for nuc_rec in nucs:
                #cov_rec['GC'] = int(rec['G']) + int(rec['C'])
                iv = Interval(chrom, int(nuc_rec['start']), int(nuc_rec['end']), data=nuc_rec)
                yield iv

    intervals = iter_ivs(nucs)

    if args.validate:
        intervals = validated_bin_intervals(intervals, args.bin_size)

    iv_flanks = iter_bin_interval_flanks(intervals, bin_size=50, lflank=args.left_flank, rflank=args.right_flank)

    # summing up
    bin_size = args.bin_size
    flank_size = args.left_flank + args.right_flank
    header = header + ('flank_gc', 'flank_nonn', 'flank_gc_ratio')
    print (*header, sep='\t')
    for chrom, recs in groupby(iv_flanks, lambda x: x.iv.contig):
        G = C = N = 0
        for rec in recs:
            # calculation by difference

            for iv in rec.adds:
                G += int(iv.data['Gs'])
                C += int(iv.data['Cs'])
                N += int(iv.data['Ns'])
            for iv in rec.removes:
                G -= int(iv.data['Gs'])
                C -= int(iv.data['Cs'])
                N -= int(iv.data['Ns'])

            iv = rec.iv
            flank_gc =  G + C
            flank_nonn = len(rec.flanks) * bin_size - N
            iv.data['flank_gc'] = flank_gc
            iv.data['flank_nonn'] = flank_nonn
            iv.data['flank_gc_ratio'] = 1. * flank_gc / flank_nonn if flank_nonn else 0.
            print (*iv.data.values(), sep='\t')


@command.add_sub
@argument('fasta')
@argument('-s', '--bin-size', type=int, default=50)
@argument('-N', '--omit-N', action='store_true', default=False)
@argument('-H', '--no-header', dest='header', action='store_false', default=True)
@argument('--chroms', nargs='*', default=None)
def fa_nuc(args):
    """
    """
    header = 'chrom start end As Cs Gs Ts Ns'.split(' ')
    bin_size = args.bin_size
    if args.chroms:
        chroms = list(line.rstrip('\r\n').split('\t')[0] for line in open(args.chrom_list))
    else:
        chroms = None
    if args.header:
        print (*header, sep='\t')
    for chrom, start, end, seq in iter_fa_seq(args.fasta, bin_size=bin_size, chroms=chroms):
        counter = Counter(seq)
        if len(seq) < bin_size:
            counter['N'] += bin_size - len(seq)

        if args.omit_N and counter['N'] == bin_size:
            continue

        print (chrom, start, end, counter['A'], counter['C'], counter['G'], counter['T'], counter['N'], sep='\t')


def iter_fa_seq(fasta, bin_size=50, chroms=None):
    fasta = Fastafile(fasta)
    if chroms is None:
        chroms = fasta.references
    for chrom in chroms:
        start = 0
        while 1:
            seq = fasta.fetch(chrom, start, start + bin_size).upper()
            if not seq:
                break
            yield chrom, start, start + bin_size, seq
            start += bin_size


@command.add_sub
@argument('bed1')
@argument('bed2')
@argument('-H', '--no-header', dest='use_header', action='store_false', default=True)
@argument('-l', '--chrom-list', required=True, help='list of chromosome to use in order')
def bed_join(args):
    if args.use_header:
        def get_bed_ivs(bed):
            bed = iter_tabs(bed)
            bed = with_header(bed, use_header=True)
            header = next(bed)
            ivs = (Interval(rec['chrom'], int(rec['start']), int(rec['end']), data=rec) for rec in bed
                    if rec['chrom'] in chrom_set)
            return header, ivs

        def get_row(iv):
            iv1 = iv.data[0]
            iv2 = iv.data[1]
            return (iv.contig, iv.start, iv.end,) + tuple(iv1[k] for k in data_key1) + tuple(iv2[k] for k in data_key2)

    else:
        def get_bed_ivs(bed):
            bed = iter_tabs(bed)
            return None, (Interval(rec[0], int(rec[1]), int(rec[2]), data=rec) for rec in bed
                            if rec[0] in chrom_set)

        def get_row(iv):
            iv1 = iv.data[0]
            iv2 = iv.data[1]
            return (iv.contig, iv.start, iv.end,) + tuple(iv1[3:]) + tuple(iv2[3:])

    header1, ivs1 = get_bed_ivs(open(args.bed1))
    header2, ivs2 = get_bed_ivs(open(args.bed2))

    if args.use_header:
        data_key1 = header1[3:]
        data_key2 = header2[3:]
        # header
        print (*(('chrom', 'start', 'end') + data_key1 + data_key2), sep='\t')

    chroms = list(line.rstrip('\r\n').split('\t')[0] for line in open(args.chrom_list))
    chrom_set = set(chroms)
    comp = ChromCompare(chroms)

    for iv in align_intervals(ivs1, ivs2, comp):
        iv1 = iv.data[0]
        iv2 = iv.data[1]
        if iv1 is None or iv2 is None:
            logging.info('%s was skipped', iv)
            continue
        row = get_row(iv)
        print (*row, sep='\t')


_re_autosomal = re.compile('^(chr[0-9]+|[0-9]+)$')

def is_autosomal(chrom):
    """
    >>> is_autosomal('chr1')
    True
    >>> is_autosomal('19')
    True
    >>> is_autosomal('22')
    True
    >>> is_autosomal('X')
    False
    >>> is_autosomal('chrY')
    False
    >>> is_autosomal('chr6_random')
    False
    """
    return bool(_re_autosomal.match(chrom))


@command.add_sub
@argument('gc_cov')
@argument('--gc-col', default='flank_gc_ratio')
@argument('--low-mapq-col', default='mapq1')
@argument('--gc-bin-size', type=float, default=.02)
@argument('-b', '--low-mapq-bound', type=float, default=.04)
@argument('--chroms', nargs='*', default=None)
def gc_cov_count(args):
    """
    Default chroms are autosomals
    {
        gc_bin_size,
        max_multimaps,
        max_multimaps,
        counts: {gc_bin: {cov: count}},
    }
    """
    logging.info('Chromosomes: %s', args.chroms)
    if args.chroms:
        chrom_set = set(args.chroms)
        def is_target_chrom(chrom):
            return chrom in chrom_set
    else:
        is_target_chrom = is_autosomal

    gc_col = args.gc_col
    low_mapq_col = args.low_mapq_col
    gc_bin_size = args.gc_bin_size
    bound = args.low_mapq_bound

    with open(args.gc_cov) as fp:
        gc_covs = iter_tabs(fp)
        gc_covs = with_header(gc_covs)

        gc_cov_counters = defaultdict(Counter)  # {gc_bin: {cov: count}}
        for chrom, gc_covs in groupby(gc_covs, lambda x: x['chrom']):
            if not is_target_chrom(chrom):
                continue

            logging.info('Processing chrom: %s', chrom)
            for gc_cov in gc_covs:
                cov = int(gc_cov['count'])
                low_cov = int(gc_cov[low_mapq_col])
                gc_ratio = float(gc_cov[gc_col])
                if cov > 0 and 1. * low_cov / cov >= args.low_mapq_bound:
                    continue  # skip low mapq record

                gc_bin = int(gc_ratio / gc_bin_size)
                gc_cov_counters[gc_bin][cov] += 1


    gc_bin_max = int(round(1. / gc_bin_size))
    print ('gc_bin', 'gc_ratio', 'cov', 'count', sep='\t')
    for gc_bin in xrange(gc_bin_max):
        for cov, count in sorted(gc_cov_counters[gc_bin].items()):
            print (gc_bin, '{0:.2f}'.format(gc_bin * gc_bin_size), cov, count, sep='\t')


_gc_stats_header = ['gc_bin', 'gc_ratio', 'nbins_total', 'cov_mean_total', 'nbins_total_trim', 'cov_mean_total_trim', 'nbins', 'cov_mean', 'cov_std', 'nbins_trim', 'cov_mean_trim']
def make_gc_stats(gc_cov_count, out=sys.stdout, min_nbins=100):
    """  # TODO use panads
    gc_bin gc_ratio gc_ratio
    """
    gc_cov_counts = defaultdict(Counter)
    cov_counts = Counter()
    gc_ratios = {}
    q1 = .1
    q2 = .9

    with open(gc_cov_count) as fp:
        for rec in with_header(iter_tabs(fp)):
            gc_bin = int(rec['gc_bin'])
            gc_ratio = float(rec['gc_ratio'])
            cov = int(rec['cov'])
            count = int(rec['count'])
            cov_counts[cov] += count
            gc_cov_counts[gc_bin][cov] += count
            gc_ratios[gc_bin] = gc_ratio

    print (*_gc_stats_header, file=out, sep='\t')

    covs_total = np.array(sorted(cov_counts.keys()))
    weights_total = np.array([cov_counts[cov] for cov in covs_total]).astype('float')
    cov_mean_total = (covs_total * weights_total).sum() / weights_total.sum()

    nbins_total = int(weights_total.sum())
    weights_total_trim = trim_weights(weights_total, q1, q2)
    nbins_total_trim = weights_total_trim.sum().astype('int')
    cov_mean_total_trim = trimmed_mean(covs_total, weights_total, q1, q2)

    logging.info('nbins_total: %s', nbins_total)
    logging.info('cov_mean_total: %s', cov_mean_total)
    logging.info('nbins_total_trim: %s', nbins_total_trim)
    logging.info('cov_mean_total_trim: %s', cov_mean_total_trim)

    for gc_bin in sorted(gc_cov_counts):
        gc_ratio = gc_ratios[gc_bin]
        covs = np.array(list(gc_cov_counts[gc_bin].keys())) * 1.
        bins = list(gc_cov_counts[gc_bin].values())
        weights = np.array(bins) * 1.
        cov_mean = np.average(covs, weights=weights)
        cov_std = np.sqrt((np.average(covs**2, weights=weights) - cov_mean**2).sum())
        #remains = (covs >= cov_mean_total * trim_rate_min) & (covs <= cov_mean_total * trim_rate_max)
        weights_trim = trim_weights(weights, q1, q2)
        nbins_trim = int(weights_trim.sum().round())
        cov_mean_trim = trimmed_mean(covs, weights, q1, q2) if nbins_trim >= min_nbins else float('nan')

        nbins = sum(bins)
        row = [gc_bin, '{0:.2f}'.format(gc_ratio),
            nbins_total, '{0:.4f}'.format(cov_mean_total),
            nbins_total_trim, '{0:.4f}'.format(cov_mean_total_trim),
            nbins, '{0:.4f}'.format(cov_mean), '{0:.4f}'.format(cov_std),
            nbins_trim, '{0:.4f}'.format(cov_mean_trim)]
        print (*row, file=out, sep='\t')


def load_gc_stats(stats_file):
    gc_stats = {}
    with open(stats_file) as fp:
        for row in with_header(iter_tabs(fp)):
            gc_bin = int(row['gc_bin'])
            gc_stats[gc_bin] = row
            for k in 'gc_bin nbins_total nbins_total_trim nbins nbins_trim'.split(' '):
                gc_stats[gc_bin][k] = safeint(row[k], default=row[k])
            for k in 'gc_ratio cov_mean_total cov_mean_total_trim cov_mean cov_std cov_mean_trim'.split(' '):
                gc_stats[gc_bin][k] = safefloat(row[k], default=row[k])

    return gc_stats


def trim_weights(weight, q1=0.1, q2=0.9):
    """ Trim weight outside of quantiles
    Args:
        weight: numpy array or pandas.Series object

    # print (hist_quantile(df['x'], df['w'], 0.5))
    >>> weight = np.array([5, 10, 20, 30, 20, 10, 5])
    >>> list(trim_weights(weight, 0.025, 0.975))
    [2.5, 10.0, 20.0, 30.0, 20.0, 10.0, 2.5]
    >>> list(trim_weights(weight, 0.05, 0.95))
    [0.0, 10.0, 20.0, 30.0, 20.0, 10.0, 0.0]
    >>> list(trim_weights(weight, 0.1, 0.9))
    [0.0, 5.0, 20.0, 30.0, 20.0, 5.0, 0.0]
    >>> list(trim_weights(weight, 0.2, 0.8))
    [0.0, 0.0, 15.0, 30.0, 15.0, 0.0, 0.0]
    >>> list(trim_weights(weight, 0.25, 0.75))
    [0.0, 0.0, 10.0, 30.0, 10.0, 0.0, 0.0]
    >>> list(trim_weights(weight, 0.4, 0.6))
    [0.0, 0.0, 0.0, 20.0, 0.0, 0.0, 0.0]
    >>> list(trim_weights(weight, 0.025, 0.03))
    [0.5, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    >>> list(trim_weights(weight, 0.97, 0.975))
    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.5]
    """
    assert q1 < q2
    assert np.all(weight > 0)
    wcum = np.array([0] + list(weight.cumsum()))
    tot = weight.sum()

    idx1 = wcum[wcum <= tot*q1].shape[0] - 1
    wsub1 = tot * q1- wcum[idx1]
    idx2 = wcum[wcum <= tot*q2].shape[0] - 1
    wsub2 = wcum[idx2] - tot * q2

    new_weight = weight.copy().astype('float')

    new_weight[:idx1] = 0
    new_weight[idx2:] = 0
    new_weight[idx1] -= wsub1
    new_weight[idx2] -= wsub2
    return new_weight

def trimmed_mean(vals, weights, q1, q2):
    w = trim_weights(weights, q1, q2)
    s = w.sum()
    return (1. * (vals * w).sum() / s) if s else float('nan')


@command.add_sub
@argument('gc_cov_count')
@argument('-o', '--output')
@argument('--gc-bin-size', type=float, default=None)
@argument('--plot-box', action='store_true')
def plot_gc_cov_count(args):
    #logging.info('Creating stats file: %s', stats_file)
    #with open(stats_file, 'w+') as out:
    #    make_gc_stats(args.gc_cov_count, out=out)
    #stats_file = '{0}.stats'.format(args.gc_cov_count)
    #stats = load_gc_stats(stats_file)
    import pandas as pd
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.cm as cm
    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_pdf import PdfPages
    try:
        import seaborn
    except ImportError:
        pass

    outname = args.output if args.output else '{0}.pdf'.format(args.gc_cov_count)
    logging.info('Make plot to %s', outname)

    def get_quantile_coverage(tab, q):
        tab = pd.DataFrame({'cov': list(tab['cov']), 'count': list(tab['count'])})   # workaround
        cov_idxs = tab['cov'].argsort()

        cum = tab['count'][cov_idxs].cumsum()
        tot = tab['count'].sum()

        last_idx = cum[1. * cum / tot <= q].last_valid_index()
        return tab['cov'][last_idx] if last_idx is not None else float('nan')

    # setup data
    tab = pd.read_table(args.gc_cov_count)

    q_cov = get_quantile_coverage(tab, .9)
    logging.info('The %s quantile of the coverage: %s', .9, q_cov)
    max_cov = safeint(q_cov * 2., default=10)

    def gen_data_or_sample(values, weights, size=100):
        if len(values) > size:
            return values[weighted_choice(weights, size)]
        else:
            return np.repeat(values, weights)
    # plotting
    pp = PdfPages(outname)

    gc_ratios = []
    cov_1s = []
    cov_3s = []
    cov_25s = []
    cov_75s = []
    cov_samples = []
    cov_trims = []
    for gc_bin, tab1 in tab.groupby('gc_bin'):
        gc_ratio = tab1['gc_ratio'][tab1.index[0]]
        gc_ratios.append(gc_ratio)
        cov_sample = gen_data_or_sample(tab1['cov'].values, tab1['count'].values)
        covs = np.sort(tab1['cov'].values)
        counts = tab1['count'].values[tab1['cov'].values.argsort()]
        #logging.info('%s', (covs, counts))
        cov_samples.append(cov_sample)
        cov_trims.append(trimmed_mean(covs, counts, .1, .9))
        #cov_25s.append(trimmed_mean(covs, counts, .1, .4))
        #cov_75s.append(trimmed_mean(covs, counts, .6, .9))
        cov_1s.append(get_quantile_coverage(tab1, .25))
        cov_3s.append(get_quantile_coverage(tab1, .75))

    #covs = pd.DataFrame({'gc_ratio': gc_ratios, 'cov_1': cov_1s, 'cov_med': cov_meds, 'cov_3': cov_3s})
    fig, ax = plt.subplots()
    ax.set_xlim((0, 1.))
    ax.set_ylim(0, max_cov)
    plt.plot(gc_ratios, cov_trims)
    plt.plot(gc_ratios, cov_1s, ls='dashed')
    plt.plot(gc_ratios, cov_3s, ls='dashed')
    pp.savefig(fig)

    if args.plot_box:
        fig, ax = plt.subplots()
        ax.set_xlim((0, 1.))
        ax.set_ylim(0, max_cov)
        ax.set_xticklabels(ax.xaxis.get_majorticklabels(), rotation=90, size=7)
        plt.boxplot(cov_samples, labels=gc_ratios)

        #covs.plot(kind='line', x='gc_ratio', ax=ax)
        pp.savefig(fig)

    #fig = plt.figure()
    fig, ax = plt.subplots()
    ax.set_xlim((0, max_cov))
    ax.set_ylim((0, 1.))
    x = np.arange(0, 1, 0.025)
    for gc_bin, tab1 in tab.groupby('gc_bin'):
        gc_ratio = tab1['gc_ratio'][tab1.index[0]]
        color = cm.rainbow(gc_ratio, 0.6)
        #logging.info('Color for gc %s : %s', gc_ratio, color)
        s = tab1['count'].sum()
        ax.plot(tab1['cov'].values, 1. * tab1['count'].values / s, c=color, label=gc_ratio)
    ax.legend(tab['gc_ratio'].unique(), loc='upper right', ncol=3, prop={'size': 9}) #bbox_to_anchor=(1.05, 1), ncol=3)
    pp.savefig(fig)

    #fig, ax = plt.subplots()
    #ax.set_xlim((0, 1.))
    #ax.set_ylabel('region count')
    ##ax.set_ylim((0, max_cov))
    ##tab.plot(kind='hexbin', x='gc_ratio', y='cov', C='count', ax=ax)
    #tab.groupby('gc_ratio')['count'].sum().plot(kind='area', ax=ax)
    #tab.plot(kind='scatter', x='gc_ratio', y='cov', ax=ax)
    #pp.savefig(fig)

    pp.close()


@command.add_sub
@argument('gc_cov_count')
@argument('--min-nbins', type=int, default=100)
def gc_cov_stats(args):
    make_gc_stats(args.gc_cov_count, min_nbins=args.min_nbins)


# deprecated
@command.add_sub
@argument('gc_cov')
@argument('-c', '--gc-cov-count', required=True)
@argument('-b', '--low-mapq-bound', type=float, default=.04)
@argument('--gc-bin-size', type=float, default=None)
@argument('--gc-col', default='flank_gc_ratio')
@argument('--min-nbins', type=int, default=100)
def gc_cov_normalize(args):
    stats_file = '{0}.stats'.format(args.gc_cov_count)
    #if not os.path.exists(stats_file) or os.path.getsize(stats_file) == 0 or os.path.getatime(stats_file) < os.path.getatime(args.gc_cov_count):
    logging.info('Creating stats file: %s', stats_file)
    if not os.path.exists(stats_file) or os.path.getatime(stats_file) < os.path.getatime(args.gc_cov_count):
        with open(stats_file, 'w+') as out:
            make_gc_stats(args.gc_cov_count, min_nbins=args.min_nbins, out=out)
    stats = load_gc_stats(stats_file)
    gc_means = dict((gc_bin, row['cov_mean_trim']) for gc_bin, row in stats.items())
    total_mean = list(stats.values())[0]['cov_mean_total_trim']
    gc_col = args.gc_col

    # estimate gc_bin_size
    if args.gc_bin_size is None:
        gc_ratio_lis = np.array(list(sorted(v['gc_ratio'] for v in stats.values())))
        gc_bin_size = np.min(np.diff(gc_ratio_lis))  # calculate from the difference of adjacent bin
    else:
        gc_bin_size = args.gc_bin_size

    logging.info('gc_bin_size: %s', gc_bin_size)

    #for gc_bin in sorted(stats):
    #    print (stats[gc_bin])

    with open(args.gc_cov) as fp:
        it = with_header(iter_tabs(fp), use_header=True)
        keys = next(it)
        header = keys + ('gc_bin', 'cov_normal', 'cov_gc_normal')
        print (*header, sep='\t')

        for row in it:
            vals = [row[k] for k in keys]
            gc_ratio = float(row[gc_col])
            gc_bin = int(round(gc_ratio / gc_bin_size))
            gc_mean = gc_means.get(gc_bin, float('nan'))
            cov_normal = safediv(float(row['count']), total_mean)
            cov_gc_normal = safediv(float(row['count']), gc_mean)

            vals.append(gc_bin)
            vals.append('{0:.4f}'.format(cov_normal))
            vals.append('{0:.4f}'.format(cov_gc_normal))
            print (*vals, sep='\t')


@command.add_sub
@argument('region', default='')
@argument('cov', default='coverage file tabixed')
@argument('-c', '--cov-col', default='cov_gc_normal', help='column name of coverage info')
@argument('-b', '--low-mapq-bound', type=float, default=.04)
@argument('--low-mapq-col', default='mapq1')
@argument('-r', '--use-region-header', action='store_true', default=False)
@argument('-s', '--stats-file', default=None, help='used for normalization')
@argument('--gc-col', default='flank_gc_ratio', help='used for normalizatoin')
@argument('--max-coeff', type=float, default=3.0)
@argument('--min-coeff', type=float, default=0.3)
def cov_intersect(args):
    """

    Output bed file:
      - 1 - region.ncol : region file record
      - extra fields:
        mean coverage
        valid length
        filtered length (due to excess lowq reads or unknown coverage)
        total length
    """
    cov_header = gzip.open(args.cov).readline().rstrip('\r\n').split('\t')
    idx_cov = cov_header.index(args.cov_col)  # column index of coverage data
    idx_count = cov_header.index('count')
    idx_low_mapq = cov_header.index(args.low_mapq_col)
    low_mapq_bound = args.low_mapq_bound
    #header_ext = ['mean_cov', 'region_len', 'valid_len', 'valid_ratio', 'raw_mean_count', 'raw_mean_low_mapq', 'raw_low_mapq_ratio']
    header_ext = ['mean_cov', 'region_len', 'valid_len', 'valid_ratio', 'valid_count', 'valid_exp_count', 'count', 'low_mapq_ratio']
    NEARLY_ZERO = 1e-10

    if args.stats_file:
        # for normalization
        idx_gc = cov_header.index(args.gc_col)
        max_coeff = args.max_coeff
        min_coeff = args.min_coeff

        stats = load_gc_stats(args.stats_file)
        total_mean = list(stats.values())[0]['cov_mean_total_trim']
        gc_covs = {}
        for gc_bin, row in stats.items():
            gc_ratio = row['gc_ratio']
            coeff = 1. * safediv(total_mean, row['cov_mean_trim'], error=float('nan'))   # strength of correction
            gc_covs[gc_bin] = row['cov_mean_trim'] if (min_coeff <= coeff <= max_coeff) else float('nan')
            logging.info('Reference coverage for bin %s, gc_ratio %s: %s (coeff: %s)', gc_bin, gc_ratio, gc_covs[gc_bin], coeff)

        # estimate gc_bin_size
        gc_ratio_lis = np.array(list(sorted(v['gc_ratio'] for v in stats.values())))
        gc_bin_size = np.min(np.diff(gc_ratio_lis))  # calculate from the difference of adjacent bin
        logging.info('gc_bin_size: %s', gc_bin_size)
        def get_cov(rec):
            gc_ratio = float(rec[idx_gc])
            gc_bin = int(round(gc_ratio / gc_bin_size))
            cov = gc_covs.get(gc_bin, float('nan'))
            return safediv(float(rec[idx_count]), cov)

        def get_exp_count(rec):
            gc_ratio = float(rec[idx_gc])
            gc_bin = int(round(gc_ratio / gc_bin_size))
            return gc_covs.get(gc_bin, float('nan'))
    else:
        def get_cov(rec):
            return float(rec[idx_cov])

        def get_exp_count(rec):
            return float('nan')

    with Tabixfile(args.cov) as cov_fp:
        with open(args.region) as reg_fp:
            regs = iter_tabs(reg_fp)

            first = next(regs, None)  # peaking first line of the region file
            if first is None:
                return
            if args.use_region_header:
                header = list(first) + header_ext
            else:
                header = ['chrom', 'start', 'end'] \
                    + ['regcol{0}'.format(i+1) for i in xrange(len(first) - 3)] + header_ext
                regs = chain([first], regs)

            print (*header, sep='\t')
            for reg_row in regs:
                chrom = reg_row[0]
                start = int(reg_row[1])
                end = int(reg_row[2])
                region_len = end - start

                # TODO unit test
                recs = list(iter_tabs(cov_fp.fetch(chrom, start, end)))
                if recs:
                    bin1_start = int(recs[0][1])  # first bin
                    bin1_end  = int(recs[0][2])
                    bin2_start = int(recs[-1][1]) # last bin
                    bin2_end  = int(recs[-1][2])
                    bin_size = bin1_end - bin1_start

                    mapqs = [float(rec[idx_low_mapq]) for rec in recs]  # mapqs of each recs
                    bin_len1 = bin_len2 = bin_size
                    if bin1_start < start:
                        if end < bin1_end:   # the bin contains the whole region
                            bin_len1 = end - start
                        else:
                            bin_len1 = bin1_end - start
                    if end < bin2_end:
                        if bin2_start < start:   # the bin contains the whole region
                            bin_len2 = end - start
                        else:
                            bin_len2 = end - bin2_start

                    exp_counts = np.array([get_exp_count(rec) for rec in recs])
                    covs = np.array([get_cov(rec) for rec in recs])
                    counts = np.array([int(cov[idx_count]) for cov in recs])
                    low_mapqs = np.array([int(cov[idx_low_mapq]) for cov in recs])
                    invalids = (np.isnan(covs) | (1. * (low_mapqs / (counts + NEARLY_ZERO)) >= low_mapq_bound))

                    bin_lens = np.array([bin_size] * len(recs))
                    bin_lens[0] = bin_len1 # weight correction for the left end bin
                    bin_lens[-1] = bin_len2 # weight correction for the right end bin

                    valid_len = bin_lens[~ invalids].sum()
                    mean_cov = (covs * bin_lens)[~ invalids].sum() / valid_len if valid_len else float('nan')
                    valid_ratio = 1. * valid_len / region_len

                    valid_count = 1. * (counts * bin_lens / bin_size)[~ invalids].sum()
                    valid_exp_count = 1. * (exp_counts * bin_lens / bin_size)[~ invalids].sum()
                    # counts[~ invalids].sum() * bin_size / valid_len if valid_len else float('nan')
                    count = 1.* counts.sum() #* bin_size / region_len
                    low_mapq_ratio = safediv(1.* low_mapqs.sum(), count) #* bin_size / region_len
                else:
                    mean_cov = float('nan')
                    valid_len = 0
                    valid_ratio = 0.
                    #valid_mean_count = 0.

                    valid_count = 0.
                    valid_exp_count = float('nan')
                    count = 0.
                    low_mapq_ratio = 0.

                out_row = list(reg_row)
                out_row.append('{0:.3f}'.format(mean_cov))
                out_row.append(region_len)
                out_row.append(valid_len)
                out_row.append('{0:.3f}'.format(valid_ratio))
                #out_row.append('{0:.3f}'.format(valid_mean_count))
                out_row.append('{0:.2f}'.format(valid_count))
                out_row.append('{0:.2f}'.format(valid_exp_count))
                out_row.append('{0:.2f}'.format(count))
                out_row.append('{0:.2f}'.format(low_mapq_ratio))
                #out_row.append('{0:.3f}'.format(1. * raw_mean_low_mapq / raw_mean_count if raw_mean_count else 0))
                print (*out_row, sep='\t')



